#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: Redis-based discovery and event notification

*Note:* This discovery package requires the ``redis`` package (available on
PyPI), and a Redis server.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Ensure that the redis package is imported and not this module
from __future__ import absolute_import

# Standard library
import logging
import math
import socket
import threading

# Redis
import redis

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Invalidate,
    Validate,
    Property,
)

# Pelix utilities
import pelix.constants

# Remote services
from pelix.remote.edef_io import EDEFWriter, EDEFReader
import pelix.remote
import pelix.remote.beans as beans

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# The module's logger
_logger = logging.getLogger(__name__)

PATTERN_FRAMEWORK_KEY = "pelix/remote/frameworks/{fw_uid}"
"""
Pattern of framework heartbeat key, with the following format entry:
* ``fw_uid``: Framework UID
"""

PATTERN_ENDPOINT_KEY = "pelix/remote/services/{fw_uid}/{endpoint_uid}"
"""
Pattern of endpoints keys, with the following format entries:
* ``fw_uid``: Framework UID
* ``endpoint_uid``: Export endpoint UID
"""


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_DISCOVERY_REDIS)
@Provides(pelix.remote.SERVICE_EXPORT_ENDPOINT_LISTENER)
@Requires("_dispatcher", pelix.remote.SERVICE_DISPATCHER)
@Requires("_registry", pelix.remote.SERVICE_REGISTRY)
@Property("_redis_host", "redis.host", "localhost")
@Property("_redis_port", "redis.port", 6379)
@Property("_redis_db", "redis.db", 0)
@Property("_redis_password", "redis.password")
@Property("_heart_delay", "heartbeat.delay", 10)
class RedisDiscovery(object):
    """
    Remote services discovery and notification using a Redis server
    """

    def __init__(self):
        """
        Sets up the component
        """
        # Endpoints registries
        self._dispatcher = None
        self._registry = None

        # Framework UID
        self._fw_uid = None

        # Hostname cache: Framework UID -> Hostname
        self._frameworks_hosts = {}

        # Redis
        self._redis_host = "localhost"
        self._redis_port = 6379
        self._redis_db = 0
        self._redis_password = None
        self._redis = None

        # PubSub thread
        self._pubsub = None
        self._pubsub_thread = None

        # Heartbeat thread
        self._heart_delay = 10
        self._heart_thread = None
        self._stop_event = threading.Event()

        # Event handlers
        self.__endpoint_handlers = {
            "set": self._handle_set,
            "del": self._handle_del,
        }

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Stop the Pub/Sub thread, if any
        if self._pubsub_thread is not None:
            self._pubsub_thread.stop()
            self._pubsub_thread.join()
            self._pubsub_thread = None

        # Close the Pub/Sub object
        self._pubsub.close()
        self._pubsub = None

        # Stop the heart beat
        self._unregister_framework()

        # Clean up
        self._frameworks_hosts.clear()
        _logger.debug("Redis discovery invalidated")

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Ensure we have valid integers
        self._redis_port = int(self._redis_port)
        self._redis_db = int(self._redis_db)

        # Get the framework UID
        self._fw_uid = context.get_property(pelix.constants.FRAMEWORK_UID)

        # Connect to Redis
        self._redis = redis.StrictRedis(
            self._redis_host,
            self._redis_port,
            self._redis_db,
            self._redis_password,
        )

        # Configure the server to send key events
        config_key = "notify-keyspace-events"
        current = set(self._redis.config_get(config_key)[config_key])
        current.update("K$sg")
        self._redis.config_set(config_key, "".join(current))

        # Start listening to events
        keyspace = "__keyspace@{0}__".format(self._redis_db)
        patterns = {
            "pelix/remote/frameworks/*": self._handle_framework_event,
            "pelix/remote/services/*": self._handle_endpoint_event,
        }

        self._pubsub = self._redis.pubsub()
        self._pubsub.psubscribe(
            **{
                ":".join((keyspace, pattern)): handler
                for pattern, handler in patterns.items()
            }
        )

        # Start the event thread
        self._pubsub_thread = self._pubsub.run_in_thread()

        # Set the discovery key
        self._stop_event.clear()
        self._register_framework()
        _logger.debug("Redis discovery validated")

    def _register_framework(self):
        """
        Registers the framework and its current services in Redis
        """
        # The framework key
        fw_key = PATTERN_FRAMEWORK_KEY.format(fw_uid=self._fw_uid)

        # The host name
        hostname = socket.gethostname()
        if hostname == "localhost":
            logging.warning(
                "Hostname is '%s': this will be a problem for "
                "multi-host remote services",
                hostname,
            )

        def heart():
            """
            Heartbeat method
            """
            # Register the framework once
            self._redis.set(
                fw_key, hostname, int(math.ceil(self._heart_delay * 1.2))
            )

            # Loop while we're up
            while not self._stop_event.wait(self._heart_delay):
                # Re-set the key with a new time to live
                self._redis.set(
                    fw_key, hostname, int(math.ceil(self._heart_delay * 1.2))
                )

            # Stop event set: delete the key immediately
            self._redis.delete(fw_key)

        # Start the heartbeat thread
        self._heart_thread = threading.Thread(
            target=heart, name="Redis Discovery Heartbeat"
        )
        self._heart_thread.daemon = True
        self._heart_thread.start()

        # Load existing services
        self._load_existing_endpoints()

        # Register all already exported services
        for endpoint in self._dispatcher.get_endpoints():
            self._register_service(endpoint)

    def _load_existing_endpoints(self):
        """
        Loads already-registered endpoints
        """
        fw_cache = {}
        for key in self._redis.keys(
            PATTERN_ENDPOINT_KEY.format(fw_uid="*", endpoint_uid="*")
        ):
            # Extract UIDs
            key = key.decode("utf-8")
            fw_uid, _ = self._extract_uids(key)

            # Check if the framework is still there
            try:
                # Check cached information
                register = fw_cache[fw_uid]
            except KeyError:
                # Check a new framework
                register = fw_cache[fw_uid] = self._redis.exists(
                    PATTERN_FRAMEWORK_KEY.format(fw_uid=fw_uid)
                )

            if register:
                # Framework exists: store the endpoint
                self._handle_set(key)
            else:
                # Framework has expired: cleanup the remaining keys
                self._redis.delete(key)

    def _unregister_framework(self):
        """
        Removes all entries related to the framework in Redis
        """
        # Stop the thread setting the entry
        self._stop_event.set()
        self._heart_thread.join()
        self._heart_thread = None

        # The framework is already unregistered: remove all the endpoints
        endpoints = self._redis.keys(
            PATTERN_ENDPOINT_KEY.format(fw_uid=self._fw_uid, endpoint_uid="*")
        )
        if endpoints:
            self._redis.delete(*endpoints)

    def _register_service(self, endpoint):
        """
        Registers/Updates an exported service to Redis.

        :param endpoint: A :class:`~pelix.remote.ExportEndpoint` object
        """
        # Convert to EDEF
        key = PATTERN_ENDPOINT_KEY.format(
            fw_uid=self._fw_uid, endpoint_uid=endpoint.uid
        )
        xml_string = EDEFWriter().to_string(
            [beans.EndpointDescription.from_export(endpoint)]
        )

        # Send to Redis (without expiration)
        self._redis.set(key, xml_string)

    def _unregister_service(self, endpoint):
        """
        Unregisters an endpoint from Redis

        :param endpoint: A :class:`~pelix.remote.ExportEndpoint` object
        """
        self._redis.delete(
            PATTERN_ENDPOINT_KEY.format(
                fw_uid=self._fw_uid, endpoint_uid=endpoint.uid
            )
        )

    def endpoints_added(self, endpoints):
        """
        Multiple endpoints have been added

        :param endpoints: A list of ExportEndpoint beans
        """
        for endpoint in endpoints:
            self._register_service(endpoint)

    def endpoint_updated(self, endpoint, _):
        """
        An end point is updated

        :param endpoint: The updated endpoint
                         (:class:`~pelix.remote.ExportEndpoint`)
        :param _: Previous end point properties (not used)
        """
        # Update and registration are the same with Redis
        self._register_service(endpoint)

    def endpoint_removed(self, endpoint):
        """
        An end point is removed

        :param endpoint: :class:`~pelix.remote.ExportEndpoint` being removed
        """
        self._unregister_service(endpoint)

    def _handle_framework_event(self, data):
        """
        Handles a Redis notification about a framework

        :param data: Redis notification data
        """
        event = data["data"].decode("utf-8")

        # Compute framework UID
        # 1: remove the Redis channel prefix
        # 2: remove the key prefix (pelix/remote/frameworks)
        fw_key = data["channel"].decode("utf-8").split(":", 1)[1]
        fw_uid = fw_key.split("/", 3)[-1]

        if event == "expired":
            # A framework has expired: clean it up
            logging.warning("Framework %s has expired", fw_uid)

            # Forget about its hostname
            try:
                del self._frameworks_hosts[fw_uid]
            except KeyError:
                pass

            # Forget all of its endpoints
            self._registry.lost_framework(fw_uid)

            # Remove all of its endpoints (many frameworks will do that)
            keys = self._redis.keys(
                PATTERN_ENDPOINT_KEY.format(fw_uid=fw_uid, endpoint_uid="*")
            )
            if keys:
                self._redis.delete(*keys)
        elif event == "expire" and fw_uid not in self._frameworks_hosts:
            # Unknown framework found: store its hostname
            hostname = self._redis.get(fw_key)
            if hostname:
                self._frameworks_hosts[fw_uid] = hostname.decode("utf-8")

    def _handle_endpoint_event(self, data):
        """
        Handles a Redis notification about an endpoint

        :param data: Redis notification data
        """
        event = data["data"].decode("utf-8")
        try:
            # Get the event method
            method = self.__endpoint_handlers[event]
        except KeyError:
            # Ignored event
            pass
        else:
            # Extract the endpoint UID and call the handler
            method(data["channel"].decode("utf-8").split(":", 1)[1])

    @staticmethod
    def _extract_uids(endpoint_key):
        """
        Extracts the framework UID and the endpoint UID from a Redis key name

        :param endpoint_key: Name of a Redis key
        :return: A (framework UID, endpoint UID) tuple
        """
        tokens = endpoint_key.split("/")

        # Framework UID can be customized: remove prefix and the endpoint UID
        # Endpoint UID is always a well formatted UUID (no slash)
        return "/".join(tokens[3:-1]), tokens[-1]

    def _handle_set(self, endpoint_key):
        """
        An endpoint has been set or updated in Redis

        :param endpoint_key: Name of the Redis key describing the endpoint
        :return: True if an endpoint was added, False if it was a pending
                 endpoint and None if it was an echo
        """
        fw_uid, _ = self._extract_uids(endpoint_key)
        if fw_uid == self._fw_uid:
            # Got an echo
            return False

        # Not an echo: handle the event
        # 1. Get the framework hostname
        try:
            # Find in cache
            hostname = self._frameworks_hosts[fw_uid]
        except KeyError:
            # Get it from Redis
            hostname = self._redis.get(
                PATTERN_FRAMEWORK_KEY.format(fw_uid=fw_uid)
            )
            if not hostname:
                # Endpoint's framework has been removed: ignore
                # (happens when two frameworks clear traces of an old one)
                logging.debug(
                    "Framework of endpoint key %s, doesn't have a hostname",
                    endpoint_key,
                )
                return False
            else:
                # Valid hostname found, convert it to a string
                hostname = self._frameworks_hosts[fw_uid] = hostname.decode(
                    "utf-8"
                )

        # 2. Read the EDEF content
        content = self._redis.get(endpoint_key)
        if not content:
            logging.debug("Endpoint description removed while handling it")
            return False

        content = content.decode("utf-8")
        for endpoint in EDEFReader().parse(content):
            # Convert to a Pelix ImportEndpoint, and set the server hostname
            endpoint = endpoint.to_import()
            endpoint.server = hostname

            if self._registry.contains(endpoint):
                # Update endpoint
                self._registry.update(endpoint.uid, endpoint.properties)
            else:
                # New endpoint
                self._registry.add(endpoint)
        return True

    def _handle_del(self, endpoint_key):
        """
        A Redis key has been deleted

        :param endpoint_key: Name of the Redis key describing the endpoint
        """
        fw_uid, endpoint_uid = self._extract_uids(endpoint_key)
        if fw_uid != self._fw_uid and self._registry.contains(endpoint_uid):
            self._registry.remove(endpoint_uid)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
EventAdmin shell commands

Provides commands to the Pelix shell to work with the EventAdmin service

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Shell constants
from pelix.shell import SERVICE_SHELL_COMMAND

# iPOPO Decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Instantiate,
)
import pelix.services

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# -----------------------------------------------------------------------------


@ComponentFactory("eventadmin-shell-commands-factory")
@Requires("_events", pelix.services.SERVICE_EVENT_ADMIN)
@Provides(SERVICE_SHELL_COMMAND)
@Instantiate("eventadmin-shell-commands")
class EventAdminCommands(object):
    """
    EventAdmin shell commands
    """

    def __init__(self):
        """
        Sets up members
        """
        # Injected services
        self._events = None

    @staticmethod
    def get_namespace():
        """
        Retrieves the name space of this command handler
        """
        return "event"

    def get_methods(self):
        """
        Retrieves the list of tuples (command, method) for this command handler
        """
        return [("send", self.send), ("post", self.post)]

    def send(self, _, topic, **kwargs):
        """
        Sends an event (blocking)
        """
        self._events.send(topic, kwargs)

    def post(self, _, topic, **kwargs):
        """
        Posts an event (asynchronous)
        """
        self._events.post(topic, kwargs)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Python JSON ↔ Java Jabsorb format converter

Jabsorb is a serialization library for Java, converting Java beans to JSON
and vice versa.

This module is compatible with the fork of Jabsorb available at
https://github.com/Thomas Calmant/cohorte-org.jabsorb.ng

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


# Standard library
import inspect
import re

try:
    # Python 2
    # pylint: disable=F0401
    import __builtin__ as builtins
except ImportError:
    # Python 3
    # pylint: disable=F0401
    import builtins

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

JSON_CLASS = "__jsonclass__"
"""
Tuple used by jsonrpclib to indicate wich Python class corresponds to its
content
"""

JAVA_CLASS = "javaClass"
"""
Dictionary key used by Jabsorb to indicate which Java class corresponds to its
content
"""

JAVA_MAPS_PATTERN = re.compile(r"java\.util\.(.*Map|Properties)")
""" Pattern to detect standard Java classes for maps """

JAVA_LISTS_PATTERN = re.compile(r"java\.util\..*List")
""" Pattern to detect standard Java classes for lists """

JAVA_SETS_PATTERN = re.compile(r"java\.util\..*Set")
""" Pattern to detect standard Java classes for sets """

# ------------------------------------------------------------------------------


class HashableDict(dict):
    """
    Small workaround because dictionaries are not hashable in Python
    """

    def __hash__(self):
        """
        Computes the hash of the dictionary
        """
        return hash("HashableDict({0})".format(sorted(self.items())))


class HashableSet(set):
    """
    Small workaround because sets are not hashable in Python
    """

    def __hash__(self):
        """
        Computes the hash of the set
        """
        return hash("HashableSet({0})".format(sorted(self)))


class HashableList(list):
    """
    Small workaround because lists are not hashable in Python
    """

    def __hash__(self):
        """
        Computes the hash of the list
        """
        return hash("HashableList({0})".format(sorted(self)))


class AttributeMap(dict):
    """
    Wraps a map to have the same behaviour between getattr and getitem
    """

    def __init__(self, *args, **kwargs):
        """
        Adds a __dict__ member to this dictionary
        """
        super(AttributeMap, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def __hash__(self):
        """
        Computes the hash of the dictionary
        """
        return hash("AttributeMap({0})".format(sorted(self.items())))


# ------------------------------------------------------------------------------


def _compute_jsonclass(obj):
    """
    Compute the content of the __jsonclass__ field for the given object

    :param obj: An object
    :return: The content of the __jsonclass__ field
    """
    # It's not a standard type, so it needs __jsonclass__
    module_name = inspect.getmodule(obj).__name__
    json_class = obj.__class__.__name__
    if module_name not in ("", "__main__"):
        json_class = "{0}.{1}".format(module_name, json_class)

    return [json_class, []]


def _is_builtin(obj):
    """
    Checks if the type of the given object is a built-in one or not

    :param obj: An object
    :return: True if the object is of a built-in type
    """
    module_ = inspect.getmodule(obj)
    if module_ in (None, builtins):
        return True

    return module_.__name__ in ("", "__main__")


def _is_converted_class(java_class):
    """
    Checks if the given Java class is one we *might* have set up
    """
    if not java_class:
        return False

    return (
        JAVA_MAPS_PATTERN.match(java_class) is not None
        or JAVA_LISTS_PATTERN.match(java_class) is not None
        or JAVA_SETS_PATTERN.match(java_class) is not None
    )


# ------------------------------------------------------------------------------


def to_jabsorb(value):
    """
    Adds information for Jabsorb, if needed.

    Converts maps and lists to a jabsorb form.
    Keeps tuples as is, to let them be considered as arrays.

    :param value: A Python result to send to Jabsorb
    :return: The result in a Jabsorb map format (not a JSON object)
    """
    # None ?
    if value is None:
        return None

    # Map ?
    elif isinstance(value, dict):
        if JAVA_CLASS in value or JSON_CLASS in value:
            if not _is_converted_class(value.get(JAVA_CLASS)):
                # Bean representation
                converted_result = {}

                for key, content in value.items():
                    converted_result[key] = to_jabsorb(content)

                try:
                    # Keep the raw jsonrpclib information
                    converted_result[JSON_CLASS] = value[JSON_CLASS]
                except KeyError:
                    pass

            else:
                # We already worked on this value
                converted_result = value

        else:
            # Needs the whole transformation
            converted_result = {JAVA_CLASS: "java.util.HashMap"}
            converted_result["map"] = map_pairs = {}
            for key, content in value.items():
                map_pairs[key] = to_jabsorb(content)

            try:
                # Keep the raw jsonrpclib information
                map_pairs[JSON_CLASS] = value[JSON_CLASS]
            except KeyError:
                pass

    # List ? (consider tuples as an array)
    elif isinstance(value, list):
        converted_result = {
            JAVA_CLASS: "java.util.ArrayList",
            "list": [to_jabsorb(entry) for entry in value],
        }

    # Set ?
    elif isinstance(value, (set, frozenset)):
        converted_result = {
            JAVA_CLASS: "java.util.HashSet",
            "set": [to_jabsorb(entry) for entry in value],
        }

    # Tuple ? (used as array, except if it is empty)
    elif isinstance(value, tuple):
        converted_result = [to_jabsorb(entry) for entry in value]

    elif hasattr(value, JAVA_CLASS):
        # Class with a Java class hint: convert into a dictionary
        class_members = {
            name: getattr(value, name)
            for name in dir(value)
            if not name.startswith("_")
        }

        converted_result = HashableDict(
            (name, to_jabsorb(content))
            for name, content in class_members.items()
            if not inspect.ismethod(content)
        )

        # Do not forget the Java class
        converted_result[JAVA_CLASS] = getattr(value, JAVA_CLASS)

        # Also add a __jsonclass__ entry
        converted_result[JSON_CLASS] = _compute_jsonclass(value)

    # Other ?
    else:
        converted_result = value

    return converted_result


def from_jabsorb(request, seems_raw=False):
    """
    Transforms a jabsorb request into a more Python data model (converts maps
    and lists)

    :param request: Data coming from Jabsorb
    :param seems_raw: Set it to True if the given data seems to already have
                      been parsed (no Java class hint). If True, the lists will
                      be kept as lists instead of being converted to tuples.
    :return: A Python representation of the given data
    """
    if isinstance(request, (tuple, set, frozenset)):
        # Special case : JSON arrays (Python lists)
        return type(request)(from_jabsorb(element) for element in request)

    elif isinstance(request, list):
        # Check if we were a list or a tuple
        if seems_raw:
            return list(from_jabsorb(element) for element in request)

        return tuple(from_jabsorb(element) for element in request)

    elif isinstance(request, dict):
        # Dictionary
        java_class = request.get(JAVA_CLASS)
        json_class = request.get(JSON_CLASS)
        seems_raw = not java_class and not json_class

        if java_class:
            # Java Map ?
            if JAVA_MAPS_PATTERN.match(java_class) is not None:
                return HashableDict(
                    (from_jabsorb(key), from_jabsorb(value))
                    for key, value in request["map"].items()
                )

            # Java List ?
            elif JAVA_LISTS_PATTERN.match(java_class) is not None:
                return HashableList(
                    from_jabsorb(element) for element in request["list"]
                )

            # Java Set ?
            elif JAVA_SETS_PATTERN.match(java_class) is not None:
                return HashableSet(
                    from_jabsorb(element) for element in request["set"]
                )

        # Any other case
        result = AttributeMap(
            (from_jabsorb(key), from_jabsorb(value, seems_raw))
            for key, value in request.items()
        )

        # Keep JSON class information as is
        if json_class:
            result[JSON_CLASS] = json_class

        return result

    elif not _is_builtin(request):
        # Bean
        for attr in dir(request):
            # Only convert public fields
            if not attr[0] == "_":
                # Field conversion
                setattr(request, attr, from_jabsorb(getattr(request, attr)))

        return request

    else:
        # Any other case
        return request



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix shell package

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

SERVICE_SHELL = "pelix.shell"
"""
Core shell service:

* register_command(ns, command, method): Registers a command in the given name
  space, executing the given method
* unregister(namespace, command): If command is given, unregisters it, else
  unregisters the whole given namespace
* execute(cmdline, stdin, stdout): Executes the given command line with the
  given input and output streams
"""

SERVICE_SHELL_COMMAND = "pelix.shell.command"
"""
Shell commands service, for auto-registration (white board pattern).

* get_namespace(): returns the name space of the handler
* get_methods(): returns a command name → method dictionary
"""

SERVICE_SHELL_UTILS = "pelix.shell.utilities"
"""
Shell utility service:

* make_table(headers, lines): to make ASCII arrays
* bundlestate_to_str(state): to get the string representation of a bundle state
"""

SERVICE_SHELL_REMOTE = "pelix.shell.remote"
"""
Remote shell service

* get_access(): returns the (host, port) tuple where the remote shell is
  waiting clients.
"""

SERVICE_SHELL_REPORT = "pelix.shell.report"
"""
Report command service: gives access to the report methods for a future reuse

* get_levels():  Returns a copy of the dictionary of levels. The key is the
  name of the report level, the value is the tuple of methods to call for that
  level. Multiple levels can call the same method. The methods take no argument
  and return a dictionary.
* to_json(dict): Converts a dictionary to JSON, replacing inconvertible values
  to their string representation.
"""

# ------------------------------------------------------------------------------
# Temporary constants, for compatibility with previous shell developments

SHELL_SERVICE_SPEC = SERVICE_SHELL
""" Compatibility constant """

SHELL_COMMAND_SPEC = SERVICE_SHELL_COMMAND
""" Compatibility constant """

SHELL_UTILS_SERVICE_SPEC = SERVICE_SHELL_UTILS
""" Compatibility constant """

REMOTE_SHELL_SPEC = SERVICE_SHELL_REMOTE
""" Compatibility constant """

# ------------------------------------------------------------------------------

FACTORY_REMOTE_SHELL = "ipopo-remote-shell-factory"
""" Name of remote shell component factory """

FACTORY_XMPP_SHELL = "ipopo-xmpp-shell-factory"
""" Name of XMPP shell component factory """



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
iPOPO instantiation waiting list

Waits for a factory to be registered, or to appear again, to instantiate its
components.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import threading

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Dict
    from pelix.framework import BundleContext
except ImportError:
    pass

# Pelix
from pelix.constants import BundleException, BundleActivator
from pelix.internals.events import ServiceEvent
from pelix.ipopo.constants import (
    IPopoEvent,
    use_ipopo,
    SERVICE_IPOPO,
    SERVICE_IPOPO_WAITING_LIST,
)

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class IPopoWaitingList(object):
    """
    iPOPO instantiation waiting list
    """

    def __init__(self, bundle_context):
        # type: (BundleContext) -> None
        """
        Sets up members

        :param bundle_context: The bundle context
        """
        # Bundle context
        self.__context = bundle_context

        # The "queue": factory name -> {component name -> properties}
        self.__queue = {}  # type: Dict[str, Dict[str, dict]]

        # Component Name -> Factory Name
        self.__names = {}  # type: Dict[str, str]

        # Some locking
        self.__lock = threading.RLock()

    def _try_instantiate(self, ipopo, factory, component):
        # type: (Any, str, str) -> None
        """
        Tries to instantiate a component from the queue. Hides all exceptions.

        :param ipopo: The iPOPO service
        :param factory: Component factory
        :param component: Component name
        """
        try:
            # Get component properties
            with self.__lock:
                properties = self.__queue[factory][component]
        except KeyError:
            # Component not in queue
            return
        else:
            try:
                # Try instantiation
                ipopo.instantiate(factory, component, properties)
            except TypeError:
                # Unknown factory: try later
                pass
            except ValueError as ex:
                # Already known component
                _logger.error("Component already running: %s", ex)
            except Exception as ex:
                # Other error
                _logger.exception("Error instantiating component: %s", ex)

    def _start(self):
        """
        Starts the instantiation queue (called by its bundle activator)
        """
        try:
            # Try to register to factory events
            with use_ipopo(self.__context) as ipopo:
                ipopo.add_listener(self)
        except BundleException:
            # Service not yet present
            pass

        # Register the iPOPO service listener
        self.__context.add_service_listener(self, specification=SERVICE_IPOPO)

    def _stop(self):
        """
        Stops the instantiation queue (called by its bundle activator)
        """
        # Unregisters the iPOPO service listener
        self.__context.remove_service_listener(self)

        try:
            # Try to register to factory events
            with use_ipopo(self.__context) as ipopo:
                ipopo.remove_listener(self)
        except BundleException:
            # Service not present anymore
            pass

    def _clear(self):
        """
        Clear all references (called by its bundle activator)
        """
        self.__names.clear()
        self.__queue.clear()
        self.__context = None

    def service_changed(self, event):
        # type: (ServiceEvent) -> None
        """
        Handles an event about the iPOPO service
        """
        kind = event.get_kind()
        if kind == ServiceEvent.REGISTERED:
            # iPOPO service registered: register to factory events
            with use_ipopo(self.__context) as ipopo:
                ipopo.add_listener(self)

    def handle_ipopo_event(self, event):
        # type: (IPopoEvent) -> None
        """
        Handles an iPOPO event

        :param event: iPOPO event bean
        """
        kind = event.get_kind()
        if kind == IPopoEvent.REGISTERED:
            # A factory has been registered
            try:
                with use_ipopo(self.__context) as ipopo:
                    factory = event.get_factory_name()

                    with self.__lock:
                        # Copy the list of components names for this factory
                        components = self.__queue[factory].copy()

                    for component in components:
                        self._try_instantiate(ipopo, factory, component)
            except BundleException:
                # iPOPO not yet started
                pass
            except KeyError:
                # No components for this new factory
                pass

    def add(self, factory, component, properties=None):
        # type: (str, str, dict) -> None
        """
        Enqueues the instantiation of the given component

        :param factory: Factory name
        :param component: Component name
        :param properties: Component properties
        :raise ValueError: Component name already reserved in the queue
        :raise Exception: Error instantiating the component
        """
        with self.__lock:
            if component in self.__names:
                raise ValueError(
                    "Component name already queued: {0}".format(component)
                )

            # Normalize properties
            if properties is None:
                properties = {}

            # Store component description
            self.__names[component] = factory
            self.__queue.setdefault(factory, {})[component] = properties

            try:
                with use_ipopo(self.__context) as ipopo:
                    # Try to instantiate the component right now
                    self._try_instantiate(ipopo, factory, component)
            except BundleException:
                # iPOPO not yet started
                pass

    def remove(self, component):
        # type: (str) -> None
        """
        Kills/Removes the component with the given name

        :param component: A component name
        :raise KeyError: Unknown component
        """
        with self.__lock:
            # Find its factory
            factory = self.__names.pop(component)
            components = self.__queue[factory]

            # Clear the queue
            del components[component]
            if not components:
                # No more component for this factory
                del self.__queue[factory]

            # Kill the component
            try:
                with use_ipopo(self.__context) as ipopo:
                    # Try to instantiate the component right now
                    ipopo.kill(component)
            except (BundleException, ValueError):
                # iPOPO not yet started or component not instantiated
                pass


# ------------------------------------------------------------------------------


@BundleActivator
class Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Constructor
        """
        self.__registration = None
        self.__service = None

    def start(self, context):
        # type: (BundleContext) -> None
        # pylint: disable=W0212
        """
        Bundle started
        """
        # Start the service
        self.__service = IPopoWaitingList(context)
        self.__service._start()

        # Register it
        self.__registration = context.register_service(
            SERVICE_IPOPO_WAITING_LIST, self.__service, {}
        )

    def stop(self, _):
        # type: (BundleContext) -> None
        # pylint: disable=W0212
        """
        Bundle stopped
        """
        # Stop the service
        self.__service._stop()

        # Unregister the service
        self.__registration.unregister()
        self.__registration = None

        # Clear it
        self.__service._clear()
        self.__service = None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix interactive shell

Provides a console interface for the Pelix shell, based on readline when
available.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import argparse
import logging
import os
import shlex
import sys
import threading

# Initialization file handler
from pelix.misc.init_handler import InitFileHandler, remove_duplicates

# Shell constants
from pelix.constants import BundleActivator
from pelix.shell import SERVICE_SHELL
from pelix.shell.beans import IOHandler, ShellSession, safe_input
import pelix.framework as pelix

# Shell completion
from pelix.shell.completion.core import completion_hints

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

PROP_INIT_FILE = "pelix.shell.console.init_file"
""" Shell script to execute before starting the console """

PROP_RUN_FILE = "pelix.shell.console.script_file"
""" Script to run as shell input """

# ------------------------------------------------------------------------------

try:
    # Set up readline if available
    import readline

    readline.parse_and_bind("tab: complete")
    readline.set_completer(None)
except ImportError:
    # Readline is missing, not critical
    readline = None

# ------------------------------------------------------------------------------


class InteractiveShell(object):
    """
    The interactive shell handler
    """

    def __init__(self, context):
        """
        Sets up the members

        :param context: The bundle context
        """
        self._context = context  # type: pelix.BundleContext
        self._shell_ref = None  # type: pelix.ServiceReference
        self._shell = None

        # Single session
        self.__session = ShellSession(IOHandler(sys.stdin, sys.stdout), {})

        # Read line cache
        self._readline_matches = []

        # Rendez-vous events
        self._lock = threading.RLock()
        self._stop_event = threading.Event()
        self._shell_event = threading.Event()

        # Try to find a shell service
        self.search_shell()

        # Register as a service listener
        self._context.add_service_listener(self, None, SERVICE_SHELL)

    def __get_ps1(self):
        """
        Gets the prompt string from the session of the shell service

        :return: The prompt string
        """
        try:
            return self.__session.get("PS1")
        except KeyError:
            return self._shell.get_ps1()

    def _readline_prompt(self):
        """
        Prompt using the readline module (no pre-flush)

        :return: The command line
        """
        sys.stdout.flush()
        return safe_input(self.__get_ps1())

    def _normal_prompt(self):
        """
        Flushes the prompt before requesting the input

        :return: The command line
        """
        sys.stdout.write(self.__get_ps1())
        sys.stdout.flush()
        return safe_input()

    def loop_input(self, on_quit=None):
        """
        Reads the standard input until the shell session is stopped

        :param on_quit: A call back method, called without argument when the
                        shell session has ended
        """
        # Start the init script
        self._run_script(
            self.__session, self._context.get_property(PROP_INIT_FILE)
        )

        # Run the script
        script_file = self._context.get_property(PROP_RUN_FILE)
        if script_file:
            self._run_script(self.__session, script_file)
        else:
            # No script: run the main loop (blocking)
            self._run_loop(self.__session)

        # Nothing more to do
        self._stop_event.set()
        sys.stdout.write("Bye !\n")
        sys.stdout.flush()
        if on_quit is not None:
            # Call a handler if needed
            on_quit()

    def _run_script(self, session, file_path):
        """
        Runs the given script file

        :param session: Current shell session
        :param file_path: Path to the file to execute
        :return: True if a file has been execute
        """
        if file_path:
            # The 'run' command returns False in case of error
            # The 'execute' method returns False if the run command fails
            return self._shell.execute('run "{0}"'.format(file_path), session)

        return None

    def _run_loop(self, session):
        """
        Runs the main input loop

        :param session: Current shell session
        """
        try:
            first_prompt = True

            # Set up the prompt
            prompt = (
                self._readline_prompt
                if readline is not None
                else self._normal_prompt
            )

            while not self._stop_event.is_set():
                # Wait for the shell to be there
                # Before Python 2.7, wait() doesn't return a result
                if self._shell_event.wait(.2) or self._shell_event.is_set():
                    # Shell present
                    if first_prompt:
                        # Show the banner on first prompt
                        sys.stdout.write(self._shell.get_banner())
                        first_prompt = False

                    # Read the next line
                    line = prompt()

                    with self._lock:
                        if self._shell_event.is_set():
                            # Execute it
                            self._shell.execute(line, session)

                        elif not self._stop_event.is_set():
                            # Shell service lost while not stopping
                            sys.stdout.write("Shell service lost.")
                            sys.stdout.flush()
        except (EOFError, KeyboardInterrupt, SystemExit):
            # Input closed or keyboard interruption
            pass

    def readline_completer(self, text, state):
        """
        A completer for the readline library
        """
        if state == 0:
            # New completion, reset the list of matches and the display hook
            self._readline_matches = []
            try:
                readline.set_completion_display_matches_hook(None)
            except AttributeError:
                pass

            # Get the full line
            full_line = readline.get_line_buffer()
            begin_idx = readline.get_begidx()

            # Parse arguments as best as we can
            try:
                arguments = shlex.split(full_line)
            except ValueError:
                arguments = full_line.split()

            # Extract the command (maybe with its namespace)
            command = arguments.pop(0)

            if begin_idx > 0:
                # We're completing after the command (and maybe some args)
                try:
                    # Find the command
                    ns, command = self._shell.get_ns_command(command)
                except ValueError:
                    # Ambiguous command: ignore
                    return None

                # Use the completer associated to the command, if any
                try:
                    configuration = self._shell.get_command_completers(
                        ns, command
                    )
                    if configuration is not None:
                        self._readline_matches = completion_hints(
                            configuration,
                            self.__get_ps1(),
                            self.__session,
                            self._context,
                            text,
                            arguments,
                        )
                except KeyError:
                    # Unknown command
                    pass

            elif "." in command:
                # Completing the command, and a name space is given
                namespace, prefix = text.split(".", 2)
                commands = self._shell.get_commands(namespace)

                # Filter methods according to the prefix
                self._readline_matches = [
                    "{0}.{1}".format(namespace, command)
                    for command in commands
                    if command.startswith(prefix)
                ]
            else:
                # Completing a command or namespace
                prefix = command

                # Default commands goes first...
                possibilities = [
                    "{0} ".format(command)
                    for command in self._shell.get_commands(None)
                    if command.startswith(prefix)
                ]

                # ... then name spaces
                namespaces = self._shell.get_namespaces()
                possibilities.extend(
                    "{0}.".format(namespace)
                    for namespace in namespaces
                    if namespace.startswith(prefix)
                )

                # ... then commands in those name spaces
                possibilities.extend(
                    "{0} ".format(command)
                    for namespace in namespaces
                    if namespace is not None
                    for command in self._shell.get_commands(namespace)
                    if command.startswith(prefix)
                )

                # Filter methods according to the prefix
                self._readline_matches = possibilities

            if not self._readline_matches:
                return None

            # Return the first possibility
            return self._readline_matches[0]

        elif state < len(self._readline_matches):
            # Next try
            return self._readline_matches[state]

        return None

    def search_shell(self):
        """
        Looks for a shell service
        """
        with self._lock:
            if self._shell is not None:
                # A shell is already there
                return

            reference = self._context.get_service_reference(SERVICE_SHELL)
            if reference is not None:
                self.set_shell(reference)

    def service_changed(self, event):
        """
        Called by Pelix when an events changes
        """
        kind = event.get_kind()
        reference = event.get_service_reference()

        if kind in (pelix.ServiceEvent.REGISTERED, pelix.ServiceEvent.MODIFIED):
            # A service matches our filter
            self.set_shell(reference)

        else:
            with self._lock:
                # Service is not matching our filter anymore
                self.clear_shell()

                # Request for a new binding
                self.search_shell()

    def set_shell(self, svc_ref):
        """
        Binds the given shell service.

        :param svc_ref: A service reference
        """
        if svc_ref is None:
            return

        with self._lock:
            # Get the service
            self._shell_ref = svc_ref
            self._shell = self._context.get_service(self._shell_ref)

            # Set the readline completer
            if readline is not None:
                readline.set_completer(self.readline_completer)

            # Set the flag
            self._shell_event.set()

    def clear_shell(self):
        """
        Unbinds the active shell service
        """
        with self._lock:
            # Clear the flag
            self._shell_event.clear()

            # Clear the readline completer
            if readline is not None:
                readline.set_completer(None)
                del self._readline_matches[:]

            if self._shell_ref is not None:
                # Release the service
                self._context.unget_service(self._shell_ref)

            self._shell_ref = None
            self._shell = None

    def stop(self):
        """
        Clears all members
        """
        # Exit the loop
        with self._lock:
            self._stop_event.set()
            self._shell_event.clear()

        if self._context is not None:
            # Unregister from events
            self._context.remove_service_listener(self)

            # Release the shell
            self.clear_shell()
            self._context = None


# ------------------------------------------------------------------------------


@BundleActivator
class Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up the members
        """
        self._context = None
        self._shell = None
        self._thread = None

    def start(self, context):
        """
        Bundle started
        """
        self._context = context
        self._shell = InteractiveShell(context)

        # Run the loop thread
        self._thread = threading.Thread(
            target=self._shell.loop_input,
            args=[self._quit],
            name="Pelix-Shell-TextConsole",
        )
        # Set the thread as a daemon, to let it be killed by the interpreter
        # once all other threads stopped.
        self._thread.daemon = True
        self._thread.start()

    def stop(self, _):
        """
        Bundle stopped
        """
        self._cleanup()
        self._context = None

    def _quit(self):
        """
        Called when the shell session has ended
        """
        # Clean up members
        self._cleanup()

        # Stop the framework
        if self._context is not None:
            self._context.get_framework().stop()

    def _cleanup(self):
        """
        Cleans up the members
        """
        if self._shell is not None:
            # Stop the shell
            self._shell.stop()

        self._thread = None
        self._shell = None


# ------------------------------------------------------------------------------


def _resolve_file(file_name):
    """
    Checks if the file exists.

    If the file exists, the method returns its absolute path.
    Else, it returns None

    :param file_name: The name of the file to check
    :return: An absolute path, or None
    """
    if not file_name:
        return None

    path = os.path.realpath(file_name)
    if os.path.isfile(path):
        return path

    return None


def make_common_parser():
    """
    Creates an argument parser (argparse module) with the options that should
    be common to all shells.

    The result can be used as a parent parser (``parents`` argument in
    ``argparse.ArgumentParser``)

    :return: An ArgumentParser object
    """
    parser = argparse.ArgumentParser(add_help=False)

    # Version number
    parser.add_argument(
        "--version",
        action="version",
        version="Pelix {0} from {1}".format(pelix.__version__, pelix.__file__),
    )

    # Framework options
    group = parser.add_argument_group("Framework options")
    group.add_argument(
        "-D",
        nargs="+",
        dest="properties",
        metavar="KEY=VALUE",
        help="Sets framework properties",
    )
    group.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Set loggers to DEBUG level",
    )

    # Initial configuration
    group = parser.add_argument_group("Initial configuration")
    group.add_argument(
        "-c",
        "--conf",
        dest="init_conf",
        metavar="FILE",
        help="Name of an initial configuration file to use "
        "(default configuration is also loaded)",
    )
    group.add_argument(
        "-C",
        "--exclusive-conf",
        dest="init_conf_exclusive",
        metavar="FILE",
        help="Name of an initial configuration file to use "
        "(without the default configuration)",
    )
    group.add_argument(
        "-e",
        "--empty-conf",
        dest="init_empty",
        action="store_true",
        help="Don't load any initial configuration",
    )

    # Initial script
    group = parser.add_argument_group("Script execution arguments")
    group.add_argument(
        "--init",
        action="store",
        dest="init_script",
        metavar="SCRIPT",
        help="Runs the given shell script before starting the console",
    )
    group.add_argument(
        "--run",
        action="store",
        dest="run_script",
        metavar="SCRIPT",
        help="Runs the given shell script then stops the framework",
    )
    return parser


def handle_common_arguments(parsed_args):
    """
    Handles the arguments defined by :meth:`~make_common_parser`

    :param parsed_args: Argument parsed with ``argparse`` (``Namespace``)
    :return: An :class:`~InitFileHandler` object
    :raise IOError: Initial or run script not found
    """
    # Setup the logger
    logging.basicConfig(
        level=logging.DEBUG if parsed_args.verbose else logging.WARNING
    )

    # Framework properties dictionary
    props = {}

    # Read the initial configuration script
    init = InitFileHandler()
    if not parsed_args.init_empty:
        if not parsed_args.init_conf_exclusive:
            # Load default configuration
            init.load()

        # Load the given configuration file
        conf_file = parsed_args.init_conf_exclusive or parsed_args.init_conf
        if conf_file:
            init.load(conf_file)

        # Normalize configuration
        init.normalize()

        # Set initial framework properties
        props.update(init.properties)

    # Compute framework properties
    for prop_def in parsed_args.properties or []:
        key, value = prop_def.split("=", 1)
        props[key] = value

    # Check initial run script(s)
    if parsed_args.init_script:
        path = props[PROP_INIT_FILE] = _resolve_file(parsed_args.init_script)
        if not path:
            raise IOError(
                "Initial script file not found: {0}".format(
                    parsed_args.init_script
                )
            )

    if parsed_args.run_script:
        # Find the file
        path = props[PROP_RUN_FILE] = _resolve_file(parsed_args.run_script)
        if not path:
            raise IOError(
                "Script file not found: {0}".format(parsed_args.run_script)
            )

    # Update the stored configuration
    init.properties.update(props)
    return init


def main(argv=None):
    """
    Entry point

    :param argv: Script arguments (None for sys.argv)
    :return: An exit code or None
    """
    # Parse arguments
    parser = argparse.ArgumentParser(
        prog="pelix.shell.console",
        parents=[make_common_parser()],
        description="Pelix Shell Console",
    )

    # Parse arguments
    args = parser.parse_args(argv)

    # Handle arguments
    init = handle_common_arguments(args)

    # Set the initial bundles
    bundles = [
        "pelix.ipopo.core",
        "pelix.shell.core",
        "pelix.shell.ipopo",
        "pelix.shell.completion.pelix",
        "pelix.shell.completion.ipopo",
        "pelix.shell.console",
    ]
    bundles.extend(init.bundles)

    # Use the utility method to create, run and delete the framework
    framework = pelix.create_framework(
        remove_duplicates(bundles), init.properties
    )
    framework.start()

    # Instantiate components
    init.instantiate_components(framework.get_bundle_context())

    try:
        framework.wait_for_stop()
    except KeyboardInterrupt:
        framework.stop()


if __name__ == "__main__":
    # Run the entry point
    sys.exit(main() or 0)



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Alias to call pelix.shell.console.main()

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

if __name__ == "__main__":
    from pelix.shell.console import main

    main()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Shell commands for the log service

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging

# Pelix
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Instantiate,
    PostRegistration,
)
from pelix.misc import LOG_SERVICE, LOG_READER_SERVICE
from pelix.shell import SERVICE_SHELL_COMMAND

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


@ComponentFactory("pelix-shell-log-factory")
@Provides(SERVICE_SHELL_COMMAND)
@Requires("_logger", LOG_SERVICE, optional=True)
@Requires("_reader", LOG_READER_SERVICE, optional=True)
@Instantiate("pelix-shell-log")
class ShellLogCommand(object):
    """
    Provides shell commands to print the content of the log service
    """

    def __init__(self):
        """
        Sets up members
        """
        self._logger = None
        self._reader = None
        self.__svc_ref = None

    @PostRegistration
    def _post_register(self, svc_ref):
        """
        Called when the service has been provided
        """
        self.__svc_ref = svc_ref

    @staticmethod
    def get_namespace():
        """
        Returns the name space of the commands
        """
        return "log"

    def get_methods(self):
        """
        Returns the methods of the shell command
        """
        return [
            ("log", self._log),
            ("debug", self._debug),
            ("info", self._info),
            ("warn", self._warning),
            ("warning", self._warning),
            ("error", self._error),
        ]

    def _log(self, session, level="WARNING", count=None):
        """
        Prints the content of the log
        """
        if self._reader is None:
            session.write_line("No LogService available.")
            return

        # Normalize arguments
        if isinstance(level, str):
            level = logging.getLevelName(level.upper())

        if not isinstance(level, int):
            level = logging.WARNING

        if count is not None:
            try:
                safe_count = int(count)
            except (TypeError, ValueError):
                safe_count = 0
        else:
            safe_count = 0

        # Filter the entries and keep the last ones only
        try:
            for entry in [
                entry
                for entry in self._reader.get_log()
                if entry.level >= level
            ][-safe_count:]:
                session.write_line(str(entry))
        except StopIteration:
            pass

    def _trace(self, session, level, words):
        """
        Logs a message using the log service

        :param session: The shell Session object
        :param level: Log level (string)
        :param words: Message to log
        """
        if self._logger is not None:
            self._logger.log(
                level,
                " ".join(str(word) for word in words),
                None,
                self.__svc_ref,
            )
        else:
            session.write_line("No LogService available.")

    def _debug(self, session, *message):
        """
        Logs a trace
        """
        self._trace(session, logging.DEBUG, message)

    def _info(self, session, *message):
        """
        Logs an informative message
        """
        self._trace(session, logging.INFO, message)

    def _warning(self, session, *message):
        """
        Logs a warning message
        """
        self._trace(session, logging.WARNING, message)

    def _error(self, session, *message):
        """
        Logs an informative message
        """
        self._trace(session, logging.ERROR, message)



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Common parser for shell implementations

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import collections
import inspect
import logging
import shlex
import string
import sys

# Pelix modules
from pelix.utilities import to_str, get_method_arguments
import pelix.shell.beans as beans

# Shell completion
# pylint: disable=W0611
from pelix.shell.completion.decorators import ATTR_COMPLETERS, CompletionInfo

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

DEFAULT_NAMESPACE = "default"
""" Default command name space: default """

# ------------------------------------------------------------------------------


def _find_assignment(arg_token):
    """
    Find the first non-escaped assignment in the given argument token.
    Returns -1 if no assignment was found.

    :param arg_token: The argument token
    :return: The index of the first assignment, or -1
    """
    idx = arg_token.find("=")
    while idx != -1:
        if idx != 0 and arg_token[idx - 1] != "\\":
            # No escape character
            return idx

        idx = arg_token.find("=", idx + 1)

    # No assignment found
    return -1


class _ArgTemplate(string.Template):
    """
    Argument string template class
    """

    idpattern = r"[_a-z\?][_a-z0-9\.]*"


def _make_args(args_list, session, fw_props):
    """
    Converts the given list of arguments into a list (args) and a
    dictionary (kwargs).
    All arguments with an assignment are put into kwargs, others in args.

    :param args_list: The list of arguments to be treated
    :param session: The current shell session
    :return: The (arg_token, kwargs) tuple.
    """
    args = []
    kwargs = {}

    for arg_token in args_list:
        idx = _find_assignment(arg_token)
        if idx != -1:
            # Assignment
            key = arg_token[:idx]
            value = arg_token[idx + 1 :]
            kwargs[key] = value
        else:
            # Direct argument
            args.append(arg_token)

    # Prepare the dictionary of variables
    variables = collections.defaultdict(str)
    variables.update(fw_props)
    variables.update(session.variables)

    # Replace variables
    args = [_ArgTemplate(arg).safe_substitute(variables) for arg in args]
    kwargs = {
        key: _ArgTemplate(value).safe_substitute(variables)
        for key, value in kwargs.items()
    }
    return args, kwargs


def _split_ns_command(cmd_token):
    """
    Extracts the name space and the command name of the given command token.

    :param cmd_token: The command token
    :return: The extracted (name space, command) tuple
    """
    namespace = None
    cmd_split = cmd_token.split(".", 1)
    if len(cmd_split) == 1:
        # No name space given
        command = cmd_split[0]
    else:
        # Got a name space and a command
        namespace = cmd_split[0]
        command = cmd_split[1]

    if not namespace:
        # No name space given: given an empty one
        namespace = ""

    # Use lower case values only
    return namespace.lower(), command.lower()


# ------------------------------------------------------------------------------


class Shell(object):
    """
    A simple shell, based on shlex.

    Allows the use of name spaces.
    """

    def __init__(self, framework, logname=None):
        """
        Sets up members

        :param framework: The Pelix Framework instance
        :param logname: Custom name for the shell logger
        """
        self._commands = {}
        self._framework = framework
        self._logger = logging.getLogger(logname or __name__)

        # Register the help command
        self.register_command(None, "help", self.print_help)
        self.register_command(None, "?", self.print_help)

        # Basic commands
        self.register_command(None, "echo", self.echo)
        self.register_command(None, "quit", self.quit)
        self.register_command(None, "exit", self.quit)

        # Variable commands
        self.register_command(None, "set", self.var_set)
        self.register_command(None, "unset", self.var_unset)

        # File commands
        self.register_command(None, "run", self.run_file)

    @staticmethod
    def get_banner():
        """
        Returns the Shell banner
        """
        return "** Shell prompt **\n"

    @staticmethod
    def get_ps1():
        """
        Returns the PS1, the basic shell prompt
        """
        return "$ "

    def register_command(self, namespace, command, method):
        """
        Registers the given command to the shell.

        The namespace can be None, empty or "default"

        :param namespace: The command name space.
        :param command: The shell name of the command
        :param method: The method to call
        :return: True if the method has been registered, False if it was
                 already known or invalid
        """
        if method is None:
            self._logger.error("No method given for %s.%s", namespace, command)
            return False

        # Store everything in lower case
        namespace = (namespace or "").strip().lower()
        command = (command or "").strip().lower()

        if not namespace:
            namespace = DEFAULT_NAMESPACE

        if not command:
            self._logger.error("No command name given")
            return False

        if namespace not in self._commands:
            space = self._commands[namespace] = {}
        else:
            space = self._commands[namespace]

        if command in space:
            self._logger.error(
                "Command already registered: %s.%s", namespace, command
            )
            return False

        space[command] = method
        return True

    def get_command_completers(self, namespace, command):
        # type: (str, str) -> CompletionInfo
        """
        Returns the completer method associated to the given command, or None

        :param namespace: The command name space.
        :param command: The shell name of the command
        :return: A CompletionConfiguration object
        :raise KeyError: Unknown command or name space
        """
        # Find the method (can raise a KeyError)
        method = self._commands[namespace][command]

        # Return the completer, if any
        return getattr(method, ATTR_COMPLETERS, None)

    def unregister(self, namespace, command=None):
        """
        Unregisters the given command. If command is None, the whole name space
        is unregistered.

        :param namespace: The command name space.
        :param command: The shell name of the command, or None
        :return: True if the command was known, else False
        """
        if not namespace:
            namespace = DEFAULT_NAMESPACE

        namespace = namespace.strip().lower()
        if namespace not in self._commands:
            self._logger.warning("Unknown name space: %s", namespace)
            return False

        if command is not None:
            # Remove the command
            command = command.strip().lower()
            if command not in self._commands[namespace]:
                self._logger.warning(
                    "Unknown command: %s.%s", namespace, command
                )
                return False

            del self._commands[namespace][command]

            # Remove the name space if necessary
            if not self._commands[namespace]:
                del self._commands[namespace]
        else:
            # Remove the whole name space
            del self._commands[namespace]

        return True

    def __find_command_ns(self, command):
        """
        Returns the name spaces where the given command named is registered.
        If the command exists in the default name space, the returned list will
        only contain the default name space.
        Returns an empty list of the command is unknown

        :param command: A command name
        :return: A list of name spaces
        """
        # Look for the spaces where the command name appears
        namespaces = []
        for namespace, commands in self._commands.items():
            if command in commands:
                namespaces.append(namespace)

        # Sort name spaces
        namespaces.sort()

        # Default name space must always come first
        try:
            namespaces.remove(DEFAULT_NAMESPACE)
            namespaces.insert(0, DEFAULT_NAMESPACE)
        except ValueError:
            # Default name space wasn't present
            pass

        return namespaces

    def get_namespaces(self):
        """
        Retrieves the list of known name spaces (without the default one)

        :return: The list of known name spaces
        """
        namespaces = list(self._commands.keys())
        namespaces.remove(DEFAULT_NAMESPACE)
        namespaces.sort()
        return namespaces

    def get_commands(self, namespace):
        """
        Retrieves the commands of the given name space. If *namespace* is None
        or empty, it retrieves the commands of the default name space

        :param namespace: The commands name space
        :return: A list of commands names
        """
        if not namespace:
            # Default name space:
            namespace = DEFAULT_NAMESPACE

        try:
            namespace.strip().lower()
            commands = list(self._commands[namespace].keys())
            commands.sort()
            return commands
        except KeyError:
            # Unknown name space
            return []

    def get_ns_commands(self, cmd_name):
        """
        Retrieves the possible name spaces and commands associated to the given
        command name.

        :param cmd_name: The given command name
        :return: A list of 2-tuples (name space, command)
        :raise ValueError: Unknown command name
        """
        namespace, command = _split_ns_command(cmd_name)
        if not namespace:
            # Name space not given, look for the commands
            spaces = self.__find_command_ns(command)
            if not spaces:
                # Unknown command
                raise ValueError("Unknown command {0}".format(command))
            else:
                # Return a sorted list of tuples
                return sorted((namespace, command) for namespace in spaces)

        # Single match
        return [(namespace, command)]

    def get_ns_command(self, cmd_name):
        """
        Retrieves the name space and the command associated to the given
        command name.

        :param cmd_name: The given command name
        :return: A 2-tuple (name space, command)
        :raise ValueError: Unknown command name
        """
        namespace, command = _split_ns_command(cmd_name)
        if not namespace:
            # Name space not given, look for the command
            spaces = self.__find_command_ns(command)
            if not spaces:
                # Unknown command
                raise ValueError("Unknown command {0}".format(command))
            elif len(spaces) > 1:
                # Multiple possibilities
                if spaces[0] == DEFAULT_NAMESPACE:
                    # Default name space has priority
                    namespace = DEFAULT_NAMESPACE

                else:
                    # Ambiguous name
                    raise ValueError(
                        "Multiple name spaces for command '{0}': {1}".format(
                            command, ", ".join(sorted(spaces))
                        )
                    )
            else:
                # Use the found name space
                namespace = spaces[0]

        # Command found
        return namespace, command

    def execute(self, cmdline, session=None):
        """
        Executes the command corresponding to the given line

        :param cmdline: Command line to parse
        :param session: Current shell session
        :return: True if command succeeded, else False
        """
        if session is None:
            # Default session
            session = beans.ShellSession(
                beans.IOHandler(sys.stdin, sys.stdout), {}
            )

        assert isinstance(session, beans.ShellSession)

        # Split the command line
        if not cmdline:
            return False

        # Convert the line into a string
        cmdline = to_str(cmdline)

        try:
            line_split = shlex.split(cmdline, True, True)
        except ValueError as ex:
            session.write_line("Error reading line: {0}", ex)
            return False

        if not line_split:
            return False

        try:
            # Extract command information
            namespace, command = self.get_ns_command(line_split[0])
        except ValueError as ex:
            # Unknown command
            session.write_line(str(ex))
            return False

        # Get the content of the name space
        space = self._commands.get(namespace, None)
        if not space:
            session.write_line("Unknown name space {0}", namespace)
            return False

        # Get the method object
        method = space.get(command, None)
        if method is None:
            session.write_line("Unknown command: {0}.{1}", namespace, command)
            return False

        # Make arguments and keyword arguments
        args, kwargs = _make_args(
            line_split[1:], session, self._framework.get_properties()
        )
        try:
            # Execute it
            result = method(session, *args, **kwargs)

            # Store the result as $?
            if result is not None:
                session.set(beans.RESULT_VAR_NAME, result)

            # 0, None are considered as success, so don't use not nor bool
            return result is not False
        except TypeError as ex:
            # Invalid arguments...
            self._logger.error(
                "Error calling %s.%s: %s", namespace, command, ex
            )
            session.write_line("Invalid method call: {0}", ex)
            self.__print_namespace_help(session, namespace, command)
            return False
        except Exception as ex:
            # Error
            self._logger.exception(
                "Error calling %s.%s: %s", namespace, command, ex
            )
            session.write_line("{0}: {1}", type(ex).__name__, str(ex))
            return False
        finally:
            # Try to flush in any case
            try:
                session.flush()
            except IOError:
                pass

    @staticmethod
    def __extract_help(method):
        """
        Formats the help string for the given method

        :param method: The method to document
        :return: A tuple: (arguments list, documentation line)
        """
        if method is None:
            return "(No associated method)"

        # Get the arguments
        arg_spec = get_method_arguments(method)

        # Ignore the session argument
        start_arg = 1

        # Compute the number of arguments with default value
        if arg_spec.defaults is not None:
            nb_optional = len(arg_spec.defaults)

            # Let the mandatory arguments as they are
            args = [
                "<{0}>".format(arg)
                for arg in arg_spec.args[start_arg:-nb_optional]
            ]

            # Add the other arguments
            for name, value in zip(
                arg_spec.args[-nb_optional:], arg_spec.defaults[-nb_optional:]
            ):
                if value is not None:
                    args.append("[<{0}>={1}]".format(name, value))
                else:
                    args.append("[<{0}>]".format(name))
        else:
            # All arguments are mandatory
            args = ["<{0}>".format(arg) for arg in arg_spec.args[start_arg:]]

        # Extra arguments
        if arg_spec.keywords:
            args.append("[<property=value> ...]")

        if arg_spec.varargs:
            args.append("[<{0} ...>]".format(arg_spec.varargs))

        # Get the documentation string
        doc = inspect.getdoc(method) or "(Documentation missing)"
        return " ".join(args), " ".join(doc.split())

    def __print_command_help(self, session, namespace, cmd_name):
        """
        Prints the documentation of the given command

        :param session: Session handler
        :param namespace: Name space of the command
        :param cmd_name: Name of the command
        """
        # Extract documentation
        args, doc = self.__extract_help(self._commands[namespace][cmd_name])

        # Print the command name, and its arguments
        if args:
            session.write_line("- {0} {1}", cmd_name, args)
        else:
            session.write_line("- {0}", cmd_name)

        # Print the documentation line
        session.write_line("\t\t{0}", doc)

    def __print_namespace_help(self, session, namespace, cmd_name=None):
        """
        Prints the documentation of all the commands in the given name space,
        or only of the given command

        :param session: Session Handler
        :param namespace: Name space of the command
        :param cmd_name: Name of the command to show, None to show them all
        """
        session.write_line("=== Name space '{0}' ===", namespace)

        # Get all commands in this name space
        if cmd_name is None:
            names = [command for command in self._commands[namespace]]
            names.sort()
        else:
            names = [cmd_name]

        first_cmd = True
        for command in names:
            if not first_cmd:
                # Print an empty line
                session.write_line("\n")

            self.__print_command_help(session, namespace, command)
            first_cmd = False

    def print_help(self, session, command=None):
        """
        Prints the available methods and their documentation, or the
        documentation of the given command.
        """
        if command:
            # Single command mode
            if command in self._commands:
                # Argument is a name space
                self.__print_namespace_help(session, command)
                was_namespace = True
            else:
                was_namespace = False

            # Also print the name of matching commands
            try:
                # Extract command name space and name
                possibilities = self.get_ns_commands(command)
            except ValueError as ex:
                # Unknown command
                if not was_namespace:
                    # ... and no name space were matching either -> error
                    session.write_line(str(ex))
                    return False
            else:
                # Print the help of the found command
                if was_namespace:
                    # Give some space
                    session.write_line("\n\n")

                for namespace, cmd_name in possibilities:
                    self.__print_namespace_help(session, namespace, cmd_name)
        else:
            # Get all name spaces
            namespaces = list(self._commands.keys())
            namespaces.remove(DEFAULT_NAMESPACE)
            namespaces.sort()
            namespaces.insert(0, DEFAULT_NAMESPACE)

            first_ns = True
            for namespace in namespaces:
                if not first_ns:
                    # Add empty lines
                    session.write_line("\n\n")

                # Print the help of all commands
                self.__print_namespace_help(session, namespace)
                first_ns = False

        return None

    @staticmethod
    def echo(session, *words):
        """
        Echoes the given words
        """
        session.write_line(" ".join(words))

    @staticmethod
    def quit(session):
        """
        Stops the current shell session (raises a KeyboardInterrupt exception)
        """
        session.write_line("Raising KeyboardInterrupt to stop main thread")
        raise KeyboardInterrupt()

    @staticmethod
    def var_set(session, **kwargs):
        """
        Sets the given variables or prints the current ones. "set answer=42"
        """
        if not kwargs:
            for name, value in session.variables.items():
                session.write_line("{0}={1}".format(name, value))
        else:
            for name, value in kwargs.items():
                name = name.strip()
                session.set(name, value)
                session.write_line("{0}={1}", name, value)

    @staticmethod
    def var_unset(session, name):
        """
        Unsets the given variable
        """
        name = name.strip()
        try:
            session.unset(name)
        except KeyError:
            session.write_line("Unknown variable: {0}", name)
            return False
        else:
            session.write_line("Variable {0} unset.", name)

        return None

    def run_file(self, session, filename):
        """
        Runs the given "script" file
        """
        try:
            with open(filename, "r") as filep:
                for lineno, line in enumerate(filep):
                    line = line.strip()
                    if not line or line.startswith("#"):
                        # Ignore comments and empty lines
                        continue

                    # Print out the executed line
                    session.write_line("[{0:02d}] >> {1}", lineno, line)

                    # Execute the line
                    if not self.execute(line, session):
                        session.write_line(
                            "Command at line {0} failed. Abandon.", lineno + 1
                        )
                        return False

                session.write_line("Script execution succeeded")
        except IOError as ex:
            session.write_line("Error reading file {0}: {1}", filename, ex)
            return False

        return None



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

Distribution Provider API

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
from threading import RLock
import logging

from pelix.constants import OBJECTCLASS, SERVICE_SCOPE
from pelix.framework import ServiceRegistration, ServiceReference, BundleContext
from pelix.ipopo.decorators import Requires, ValidateComponent, Invalidate
from pelix.ipopo.constants import (
    SERVICE_IPOPO,
    IPOPO_INSTANCE_NAME,
    ARG_BUNDLE_CONTEXT,
    ARG_PROPERTIES,
)

# Typing
try:
    from typing import Any, Dict, List, Tuple, Optional, Callable
except ImportError:
    pass

from pelix.rsa import ImportRegistration

from pelix.rsa import (
    get_dot_properties,
    SERVICE_INTENTS,
    merge_dicts,
    ECF_RSVC_ID,
    RemoteServiceError,
    copy_non_reserved,
    ECF_SERVICE_EXPORTED_ASYNC_INTERFACES,
    ENDPOINT_ID,
    ENDPOINT_FRAMEWORK_UUID,
    SERVICE_ID,
    SERVICE_IMPORTED,
    SERVICE_IMPORTED_CONFIGS,
    REMOTE_CONFIGS_SUPPORTED,
    SERVICE_BUNDLE_ID,
    convert_string_plus_value,
    create_uuid,
    SERVICE_REMOTE_SERVICE_ADMIN,
    SERVICE_EXPORTED_CONFIGS,
    get_string_plus_property,
    get_string_plus_property_value,
    SERVICE_EXPORTED_INTENTS_EXTRA,
    SERVICE_EXPORTED_INTENTS,
)

import pelix.rsa as rsa
from pelix.rsa.endpointdescription import EndpointDescription

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

# Standard service property that is added to the set of properties provided
# to the call to ipopo.instantiate(container_factory,container_id,properties
# The property value is guaranteed to refer to the self instance of the
# distribution provider that is creating/instantiating the container
DISTRIBUTION_PROVIDER_CONTAINER_PROP = "pelix.rsa.distributionprovider"

# ------------------------------------------------------------------------------
# Abstract DistributionProvider superclass


@Requires("_rsa", SERVICE_REMOTE_SERVICE_ADMIN)
@Requires("_ipopo", SERVICE_IPOPO)
class DistributionProvider(object):
    """
    Abstract super class for all distribution providers.
    Does not expose and 'public' methods (all methods _)
    that are intended to allow subclasses (e.g. ImportDistributionProvider
    and ExportDistributionProvider to get behavior necessary
    to implement supports_import and supports_export in a manner
    consistent with OSGi RSA requirements.

    Note that subclasses should set the following fields to non-None
    values via Property annotations or in __init__:

    self._config_name  (string)
    self._namespace  (string)
    self._supported_configs  (list(string))
    self._supported_intents  (list(string))

    These two members will be set via Requires class annotations:

    self._rsa  (SERVICE_REMOTE_SERVICE_ADMIN instance)
    self._ipopo (SERVICE_IPOPO instance)
    """

    def __init__(self):
        self._config_name = None
        self._namespace = None
        self._allow_reuse = True
        self._auto_create = True
        self._supported_configs = []
        self._supported_intents = None
        self._rsa = None
        self._ipopo = None

    def get_config_name(self):
        # type: () -> str
        return self._config_name

    def get_supported_configs(self):
        # type: () -> List[str]
        return self._supported_configs

    def get_supported_intents(self):
        # type: () -> List[str]
        return self._supported_intents

    def _get_imported_configs(self, exported_configs):
        # type: (List[str]) -> List[str]
        # pylint: disable=W0613
        """
        Get any imported configs (list) given a set of exported_configs.
        Default implementation simply returns [self._config_name]
        """
        return [self._config_name]

    @staticmethod
    def _match_intents_supported(intents, supported_intents):
        # type: (List[str], List[str]) -> bool
        """
        Match the list of given intents with given supported_intents.
        This method is used by the other _match methods.
        """
        if intents is None or not supported_intents:
            return False

        return len([x for x in intents if x in supported_intents]) == len(
            intents
        )

    def _match_required_configs(self, required_configs):
        """
        Match required configs list(string).
        Default implementation compares required configs with
        self._supported_configs to make sure that all required configs are
        present for this distribution provider.
        """
        if not required_configs:
            return True
        if not self._supported_configs:
            return False
        return len(
            [x for x in required_configs if x in self._supported_configs]
        ) == len(required_configs)

    def _match_intents(self, intents):
        # type: (List[str]) -> bool
        """
        Match list(string) of intents against self._supported_intents.
        Default implementation compares intents against self._supported_intents
        """
        return self._match_intents_supported(intents, self._supported_intents)

    def _find_container(self, container_id, container_props):
        # type: (str, Dict[str, Any]) -> Any
        # pylint: disable=W0212
        """
        Uses given container_id to get an ipopo instance with name=container_id.
        If instance is returned from ipopo.get_instance(container_id), then
        instance._match_container_props(container_props) is true, then
        the instance is returned, else None
        """
        try:
            instance = self._ipopo.get_instance(container_id)
            if instance and instance._match_container_props(container_props):
                return instance
        except KeyError:
            pass

        return None

    def _prepare_container_id(self, container_props):
        # pylint: disable=R0201, W0613
        """
        Prepare and return a (string) container id.  This method is called by
        self._get_or_create_container to create an id prior to instantiating
        and instance of the appropriate Container with the instance.name set
        to the container_id returned from this method.
        This method must be overridden by subclasses as the default
        implementation raises an Exception.
        If it returns None, then no container will be created.
        """
        raise Exception(
            "DistributionProvider._prepare_container_id must be implemented "
            "by distribution provider"
        )

    def _prepare_container_props(self, service_intents, export_props):
        """
        Prepare container props (dict).
        Creates dict of props subsequently passed to
        ipopo.instantiate(factory,container_name,container_props).
        default implementation copies . properties (as per OSGi spec)
        along with service.intents and <intent>. properties.
        Also sets DISTRIBUTION_PROVIDER_CONTAINER_PROP to self.  This
        is required by Container._get_distribution_provider()
        """
        container_props = {DISTRIBUTION_PROVIDER_CONTAINER_PROP: self}
        # first get . properties for this config
        container_props.update(
            get_dot_properties(self._config_name, export_props, True)
        )
        # then add any service intents
        if service_intents != None:
            container_props[SERVICE_INTENTS] = service_intents
            for intent in service_intents:
                container_props = merge_dicts(
                    container_props,
                    get_dot_properties(intent, export_props, False),
                )
        return container_props

    def _get_or_create_container(
        self, required_configs, service_intents, all_props
    ):
        container = None
        if self._match_required_configs(
            required_configs
        ) and self._match_intents(service_intents):
            container_props = self._prepare_container_props(
                service_intents, all_props
            )
            if container_props:
                container_id = self._prepare_container_id(container_props)
                container = self._find_container(container_id, container_props)
                if not container:
                    container = self._ipopo.instantiate(
                        self._config_name, container_id, container_props
                    )
                    assert container.is_valid()
        return container

    def _find_import_registration(self, ed):
        # type: (EndpointDescription) -> Optional[ImportRegistration]
        # pylint: disable=W0212
        """
        Looks for the Import Registration matching the given endpoint
        description

        :param ed: An endpoint description
        :return: The matching ImportRegistration or None
        """
        if not ed:
            return None

        import_regs = self._rsa._get_import_regs()
        if import_regs:
            for import_reg in import_regs:
                if import_reg.match_ed(ed):
                    return import_reg

        return None

    def _handle_import(self, ed):
        # type: (EndpointDescription) -> ImportRegistration
        """
        Handle the import of an endpoint

        :param ed: An endpoint description
        :return: An import registration
        """
        return self._rsa.import_service(ed)

    def _handle_import_update(self, ed):
        # type: (EndpointDescription) -> None
        """
        Handle the update of an endpoint

        :param ed: An endpoint description
        """
        import_reg = self._find_import_registration(ed)
        if import_reg:
            import_ref = import_reg.get_import_reference()
            if import_ref:
                import_ref.update(ed)

    def _handle_import_close(self, ed):
        # type: (EndpointDescription) -> None
        """
        Cleans up import registration

        :param ed: An endpoint description
        """
        import_reg = self._find_import_registration(ed)
        if import_reg:
            import_reg.close()


# ------------------------------------------------------------------------------
# Specification for SERVICE_EXPORT_DISTRIBUTION_PROVIDER
SERVICE_EXPORT_DISTRIBUTION_PROVIDER = "pelix.rsa.exportdistributionprovider"
# Abstract implementation of SERVICE_EXPORT_DISTRIBUTION_PROVIDER extends
# DistributionProvider superclass


class ExportDistributionProvider(DistributionProvider):
    """
    Export distribution provider.

    Implements supports_export, which is called by RSA during export_service
    to give self the ability to provide a container for exporting
    the remote service described by exported_configs, service_intents, and
    export_props.

    Note:  Subclasses MUST implement/override _prepare_container_id method
    and return a string to identify the created container instance.
    """

    def supports_export(self, exported_configs, service_intents, export_props):
        """
        Method called by rsa.export_service to ask if this
        ExportDistributionProvider supports export for given
        exported_configs (list), service_intents (list), and
        export_props (dict).

        If a ExportContainer instance is returned then it is used to export
        the service.  If None is returned, then this distribution provider will
        not be used to export the service.

        The default implementation returns self._get_or_create_container.
        """
        return self._get_or_create_container(
            exported_configs, service_intents, export_props
        )


# ------------------------------------------------------------------------------
# Specification for SERVICE_IMPORT_DISTRIBUTION_PROVIDER
SERVICE_IMPORT_DISTRIBUTION_PROVIDER = "pelix.rsa.importdistributionprovider"


class ImportDistributionProvider(DistributionProvider):
    """
    Abstract implementation of SERVICE_EXPORT_DISTRIBUTION_PROVIDER
    extends DistributionProvider superclass
    """

    def _prepare_container_id(self, container_props):
        """
        Default for import containers creates a UUID for the created container.
        """
        return create_uuid()

    def supports_import(
        self, exported_configs, service_intents, endpoint_props
    ):
        """
        Method called by rsa.export_service to ask if this
        ImportDistributionProvider supports import for given
        exported_configs (list), service_intents (list), and
        export_props (dict).

        If a ImportContainer instance is returned then it is used to import
        the service.  If None is returned, then this distribution provider will
        not be used to import the service.

        The default implementation returns self._get_or_create_container.
        """
        return self._get_or_create_container(
            exported_configs, service_intents, endpoint_props
        )


# ------------------------------------------------------------------------------


class Container:
    """
    Abstract Container type supporting both ImportContainer and ExportContainer
    """

    def __init__(self):
        self._bundle_context = None  # type: BundleContext
        self._container_props = None  # type: Dict[str, Any]
        self._exported_services = {}  # type: Dict[str, Tuple[Any, EndpointDescription]]
        self._exported_instances_lock = RLock()

    def get_id(self):
        # type: () -> str
        """
        Returns the ID of this container (its component name)

        :return: The ID of this containers
        """
        return self._container_props.get(IPOPO_INSTANCE_NAME, None)

    def is_valid(self):
        # type: () -> bool
        """
        Checks if the component is valid

        :return: Always True if it doesn't raise an exception
        :raises AssertionError: Invalid properties
        """
        assert self._bundle_context
        assert self._container_props is not None
        assert self._get_distribution_provider()
        assert self.get_config_name()
        assert self.get_namespace()
        return True

    @ValidateComponent(ARG_BUNDLE_CONTEXT, ARG_PROPERTIES)
    def _validate_component(self, bundle_context, container_props):
        # type: (BundleContext, Dict[str, Any]) -> None
        """
        Component validated

        :param bundle_context: Bundle context
        :param container_props: Instance properties
        :raises AssertionError: Invalid properties
        """
        self._bundle_context = bundle_context
        self._container_props = container_props
        self.is_valid()

    @Invalidate
    def _invalidate_component(self, _):
        # type: (BundleContext) -> None
        """
        Component invalidated

        :param bundle_context: The bundle context
        """
        with self._exported_instances_lock:
            self._exported_services.clear()

    def _get_bundle_context(self):
        # type: () -> BundleContext
        """
        Returns the bundle context of this container

        :return: A bundle context
        """
        return self._bundle_context

    def _add_export(self, ed_id, inst):
        # type: (str, Tuple[Any, EndpointDescription]) -> None
        """
        Keeps track of an exported service

        :param ed_id: ID of the endpoint description
        :param inst: A tuple: (service instance, endpoint description)
        """
        with self._exported_instances_lock:
            self._exported_services[ed_id] = inst

    def _remove_export(self, ed_id):
        # type: (str) -> Optional[Tuple[Any, EndpointDescription]]
        """
        Cleans up an exported service

        :param ed_id: ID of the endpoint description
        :return: The stored tuple (service instance, endpoint description)
                 or None
        """
        with self._exported_instances_lock:
            return self._exported_services.pop(ed_id, None)

    def _get_export(self, ed_id):
        # type: (str) -> Optional[Tuple[Any, EndpointDescription]]
        """
        Get the details of an exported service

        :param ed_id: ID of an endpoint description
        :return: The stored tuple (service instance, endpoint description)
                 or None
        """
        with self._exported_instances_lock:
            return self._exported_services.get(ed_id, None)

    def _find_export(self, func):
        # type: (Callable[[Tuple[Any, EndpointDescription]], bool]) -> Optional[Tuple[Any, EndpointDescription]]
        """
        Look for an export using the given lookup method

        The lookup method must accept a single parameter, which is a tuple
        containing a service instance and endpoint description.

        :param func: A function to look for the excepted export
        :return: The found tuple or None
        """
        with self._exported_instances_lock:
            for val in self._exported_services.values():
                if func(val):
                    return val

            return None

    def _get_distribution_provider(self):
        # type: () -> DistributionProvider
        """
        Returns the distribution provider associated to this container

        :return: A distribution provider
        """
        return self._container_props[DISTRIBUTION_PROVIDER_CONTAINER_PROP]

    def get_config_name(self):
        # type: () -> str
        """
        Returns the configuration name handled by this container

        :return: A configuration name
        """
        # pylint: disable=W0212
        return self._get_distribution_provider()._config_name

    def get_namespace(self):
        # type: () -> str
        """
        Returns the namespace of this container

        :return: A namespace ID
        """
        # pylint: disable=W0212
        return self._get_distribution_provider()._namespace

    def _match_container_props(self, container_props):
        # type: (Dict[str, Any]) -> bool
        # pylint: disable=R0201, W0613
        return True

    def get_connected_id(self):
        # pylint: disable=R0201
        return None


# ------------------------------------------------------------------------------#
# Service specification for SERVICE_EXPORT_CONTAINER
SERVICE_EXPORT_CONTAINER = "pelix.rsa.exportcontainer"


class ExportContainer(Container):
    """
    Abstract implementation of SERVICE_EXPORT_CONTAINER service specification
    extends Container class. New export distribution containers should use this
    class as a superclass to inherit required behavior.
    """

    def _get_supported_intents(self):
        # type: () -> List[str]
        return self._get_distribution_provider().get_supported_intents()

    def _export_service(self, svc, ed):
        # type: (Any, EndpointDescription) -> None
        """
        Registers a service export

        :param svc: Service instance
        :param ed: Endpoint description
        """
        self._add_export(ed.get_id(), (svc, ed))

    def _update_service(self, ed):
        # type: (EndpointDescription) -> Any
        # do nothing by default, subclasses may override
        pass

    def _unexport_service(self, ed):
        # type: (EndpointDescription) -> Optional[Tuple[Any, EndpointDescription]]
        """
        Clears a service export

        :param ed: The endpoint description
        :return: The service instance and endpoint description or None
        """
        return self._remove_export(ed.get_id())

    def prepare_endpoint_props(self, intfs, svc_ref, export_props):
        # type: (List[str], ServiceReference, Dict[str, Any]) -> Dict[str, Any]
        """
        Sets up the properties of an endpoint

        :param intfs: Specifications to export
        :param svc_ref: Reference of the exported service
        :param export_props: Export properties
        :return: The properties of the endpoint
        """
        pkg_vers = rsa.get_package_versions(intfs, export_props)
        exported_configs = get_string_plus_property_value(
            svc_ref.get_property(SERVICE_EXPORTED_CONFIGS)
        )
        if not exported_configs:
            exported_configs = [self.get_config_name()]
        service_intents = set()
        svc_intents = export_props.get(SERVICE_INTENTS, None)
        if svc_intents:
            service_intents.update(svc_intents)
        svc_exp_intents = export_props.get(SERVICE_EXPORTED_INTENTS, None)
        if svc_exp_intents:
            service_intents.update(svc_exp_intents)
        svc_exp_intents_extra = export_props.get(
            SERVICE_EXPORTED_INTENTS_EXTRA, None
        )
        if svc_exp_intents_extra:
            service_intents.update(svc_exp_intents_extra)

        rsa_props = rsa.get_rsa_props(
            intfs,
            exported_configs,
            self._get_supported_intents(),
            svc_ref.get_property(SERVICE_ID),
            export_props.get(ENDPOINT_FRAMEWORK_UUID),
            pkg_vers,
            list(service_intents),
        )
        ecf_props = rsa.get_ecf_props(
            self.get_id(),
            self.get_namespace(),
            rsa.get_next_rsid(),
            rsa.get_current_time_millis(),
        )
        extra_props = rsa.get_extra_props(export_props)
        merged = rsa.merge_dicts(rsa_props, ecf_props, extra_props)
        # remove service.bundleid
        merged.pop(SERVICE_BUNDLE_ID, None)
        # remove service.scope
        merged.pop(SERVICE_SCOPE, None)
        return merged

    def export_service(self, svc_ref, export_props):
        # type: (ServiceReference, Dict[str, Any]) -> EndpointDescription
        """
        Exports the given service

        :param svc_ref: Reference to the service to export
        :param export_props: Export properties
        :return: The endpoint description
        """
        ed = EndpointDescription.fromprops(export_props)
        self._export_service(
            self._get_bundle_context().get_service(svc_ref), ed
        )
        return ed

    def update_service(self, ed):
        # type: (EndpointDescription) -> Any
        """
        Updates a service with a new endpoint description

        :param ed: The new state of the endpoint description
        :return:
        """
        return self._update_service(ed)

    def unexport_service(self, ed):
        # type: (EndpointDescription) -> Optional[Tuple[Any, EndpointDescription]]
        """
        Clears a service export

        :param ed: The endpoint description
        :return: The service instance and endpoint description or None
        """
        return self._unexport_service(ed)

    def _dispatch_exported(self, rs_id, method_name, params):
        # first lookup service instance by comparing the rs_id against the
        # service's remote service id
        service = self._find_export(
            lambda val: val[1].get_remoteservice_id()[1] == int(rs_id)
        )
        if not service:
            raise RemoteServiceError(
                "Unknown service with rs_id={0} for method call={1}".format(
                    rs_id, method_name
                )
            )
        # Get the method
        method_ref = getattr(service[0], method_name, None)
        if method_ref is None:
            raise RemoteServiceError("Unknown method {0}".format(method_name))
        # Call it (let the errors be propagated)
        if isinstance(params, (list, tuple)):
            return method_ref(*params)

        return method_ref(**params)

    def get_connected_id(self):
        # type: () -> str
        """
        Returns the ID of this container

        :return: The ID of the container
        """
        return self.get_id()


# ------------------------------------------------------------------------------#
# Service specification for SERVICE_IMPORT_CONTAINER
SERVICE_IMPORT_CONTAINER = "pelix.rsa.importcontainer"


class ImportContainer(Container):
    """
    Abstract implementation of SERVICE_IMPORT_CONTAINER service specification
    extends Container class.  New import container classes should
    subclass this ImportContainer class to inherit necessary functionality.
    """

    def _get_imported_configs(self, exported_configs):
        # type: (List[str]) -> List[str]
        # pylint: disable=W0212
        return self._get_distribution_provider()._get_imported_configs(
            exported_configs
        )

    def _prepare_proxy_props(self, endpoint_description):
        # pylint: disable=R0201
        # type: (EndpointDescription) -> Dict[str, Any]
        result_props = copy_non_reserved(
            endpoint_description.get_properties(), dict()
        )
        # remove these props
        result_props.pop(OBJECTCLASS, None)
        result_props.pop(SERVICE_ID, None)
        result_props.pop(SERVICE_BUNDLE_ID, None)
        result_props.pop(SERVICE_SCOPE, None)
        result_props.pop(IPOPO_INSTANCE_NAME, None)
        intents = endpoint_description.get_intents()
        if intents:
            result_props[SERVICE_INTENTS] = intents
        result_props[SERVICE_IMPORTED] = True
        result_props[
            SERVICE_IMPORTED_CONFIGS
        ] = endpoint_description.get_imported_configs()
        result_props[ENDPOINT_ID] = endpoint_description.get_id()
        result_props[
            ENDPOINT_FRAMEWORK_UUID
        ] = endpoint_description.get_framework_uuid()
        async_ = endpoint_description.get_async_interfaces()
        if async_:
            result_props[ECF_SERVICE_EXPORTED_ASYNC_INTERFACES] = async_
        return result_props

    def _prepare_proxy(self, endpoint_description):
        # type: (EndpointDescription) -> Any
        # pylint: disable=R0201, W0613
        raise Exception(
            "ImportContainer._prepare_proxy must be implemented by subclass"
        )

    def import_service(self, endpoint_description):
        # type: (EndpointDescription) -> ServiceRegistration
        endpoint_description.update_imported_configs(
            self._get_imported_configs(
                endpoint_description.get_remote_configs_supported()
            )
        )
        proxy = self._prepare_proxy(endpoint_description)
        if proxy:
            return self._get_bundle_context().register_service(
                endpoint_description.get_interfaces(),
                proxy,
                self._prepare_proxy_props(endpoint_description),
            )

        return None

    def unimport_service(self, endpoint_description):
        # type: (EndpointDescription) -> None
        pass



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Remote Service Admin Shell Commands

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from threading import RLock
from traceback import print_exception
import os

try:
    # pylint: disable=W0611
    from typing import Any, Tuple, List, Callable, Optional
    from pelix.framework import BundleContext
    from pelix.rsa.remoteserviceadmin import (
        ImportRegistration,
        ExportRegistration,
    )
    from pelix.rsa.providers.distribution import Container, DistributionProvider
    from pelix.shell.beans import ShellSession
except ImportError:
    pass

from pelix.constants import SERVICE_ID
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Instantiate,
    Validate,
    Requires,
    Property,
    BindField,
    UnbindField,
    Invalidate,
)
from pelix.shell import SERVICE_SHELL_COMMAND, SERVICE_SHELL_UTILS

from pelix.rsa import (
    SERVICE_REMOTE_SERVICE_ADMIN,
    SERVICE_EXPORTED_CONFIGS,
    SERVICE_EXPORTED_INTERFACES,
    prop_dot_suffix,
)
from pelix.rsa.edef import EDEFReader, EDEFWriter
from pelix.rsa.remoteserviceadmin import RemoteServiceAdminEvent

from pelix.rsa.providers.distribution import (
    SERVICE_IMPORT_CONTAINER,
    SERVICE_EXPORT_CONTAINER,
    SERVICE_IMPORT_DISTRIBUTION_PROVIDER,
    SERVICE_EXPORT_DISTRIBUTION_PROVIDER,
)

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


def _full_class_name(obj):
    """
    Returns the full name of the class of the given object

    :param obj: Any Python object
    :return: The full name of the class of the object (if possible)
    """
    module = obj.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return obj.__class__.__name__
    return module + "." + obj.__class__.__name__


RSA_COMMAND_NAME_PROP = "rsa.command"
RSA_COMMAND_FILENAME_PROP = "edeffilename"
RSA_COMMAND_EXPORT_CONFIG_PROP = "defaultexportconfig"

# ------------------------------------------------------------------------------


@ComponentFactory("rsa-command-factory")
@Requires("_utils", SERVICE_SHELL_UTILS)
@Requires("_rsa", SERVICE_REMOTE_SERVICE_ADMIN)
@Requires("_imp_containers", SERVICE_IMPORT_CONTAINER, True, True)
@Requires("_exp_containers", SERVICE_EXPORT_CONTAINER, True, True)
@Requires(
    "_imp_dist_providers", SERVICE_IMPORT_DISTRIBUTION_PROVIDER, True, True
)
@Requires(
    "_exp_dist_providers", SERVICE_EXPORT_DISTRIBUTION_PROVIDER, True, True
)
@Provides([SERVICE_SHELL_COMMAND])
@Property(
    "_edef_filename",
    prop_dot_suffix(RSA_COMMAND_NAME_PROP, RSA_COMMAND_FILENAME_PROP),
    "edef.xml",
)
@Property(
    "_export_config",
    prop_dot_suffix(RSA_COMMAND_NAME_PROP, RSA_COMMAND_EXPORT_CONFIG_PROP),
    "ecf.xmlrpc.server",
)
@Instantiate("rsa-command")
class RSACommandHandler(object):
    """
    RSA implementation of command handler service...e.g. SERVICE_SHELL_COMMAND
    Exposes a number of shell commands for RSA operations...e.g. exportservice
    importservice, listconfigs, listproviders, listcontainers, etc
    """

    SHELL_NAMESPACE = "rsa"
    CONTAINER_FORMAT = (
        "ID={0}\n\tNamespace={1}\n\tClass={2}\n\tConnectedTo={3}\n"
        "\tConnectNamespace={4}\n\tConfig Type/Distribution Provider={5}"
    )
    PROVIDER_FORMAT = "ID={0}\n\tSupported Configs={1}\n\tSupportedIntents={2}"

    def __init__(self):
        self._context = None  # type: BundleContext
        self._utils = None
        self._rsa = None
        self._imp_containers = []
        self._exp_containers = []
        self._imp_dist_providers = []
        self._exp_dist_providers = []
        self._edef_filename = None
        self._export_config = None
        self._bind_lock = RLock()

    def _bind_lists(self, field, service):
        # type: (List[Any], Any) -> None
        """
        Thread-safe handling of addition in a list of bound services

        :param field: Name of the injected field
        :param service: Injected service
        """
        with self._bind_lock:
            field.append(service)

    def _unbind_lists(self, field, service):
        # type: (List[Any], Any) -> None
        """
        Thread-safe handling of removal in a list of bound services

        :param field: Name of the injected field
        :param service: Removed service
        """
        with self._bind_lock:
            try:
                field.remove(service)
            except ValueError:
                pass

    @BindField("_imp_containers")
    def _bind_imp_containers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._bind_lists(self._imp_containers, service)

    @UnbindField("_imp_containers")
    def _unbind_imp_containers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._unbind_lists(self._imp_containers, service)

    @BindField("_exp_containers")
    def _bind_exp_containers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._bind_lists(self._exp_containers, service)

    @UnbindField("_exp_containers")
    def _unbind_exp_containers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._unbind_lists(self._exp_containers, service)

    @BindField("_imp_dist_providers")
    def _bind_imp_dist_providers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._bind_lists(self._imp_dist_providers, service)

    @UnbindField("_imp_dist_providers")
    def _unbind_imp_dist_providers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._unbind_lists(self._imp_dist_providers, service)

    @BindField("_exp_dist_providers")
    def _bind_exp_dist_providers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._bind_lists(self._exp_dist_providers, service)

    @UnbindField("_exp_dist_providers")
    def _unbind_exp_dist_providers(self, field, service, service_ref):
        # pylint: disable=W0613
        self._unbind_lists(self._exp_dist_providers, service)

    def _get_containers(self, container_id=None):
        # type: (Optional[str]) -> List[Container]
        """
        Gets the list of import and export containers

        :param container_id: An optional container ID
        :return: All containers or those matching the given ID
        """
        with self._bind_lock:
            containers = set(self._imp_containers + self._exp_containers)
            if container_id:
                return [c for c in containers if c.get_id() == container_id]

            return list(containers)

    def _get_dist_providers(self, provider_id=None):
        # type: (Optional[str]) -> List[DistributionProvider]
        """
        Gets the list of import and export providers

        :param provider_id: An optional provider ID
        :return: All providers or those matching the given ID
        """
        with self._bind_lock:
            providers = set(self._imp_dist_providers + self._exp_dist_providers)
            if provider_id:
                return [
                    p for p in providers if p.get_config_name() == provider_id
                ]

            return list(providers)

    @Validate
    def _validate(self, bundle_context):
        # type: (BundleContext) -> None
        """
        Component validated
        """
        self._context = bundle_context

    @Invalidate
    def _invalidate(self, _):
        # type: (BundleContext) -> None
        """
        Component invalidated
        """
        if self._edef_filename and os.path.isfile(self._edef_filename):
            os.remove(self._edef_filename)

    @staticmethod
    def get_namespace():
        """
        Returns the namespace of the shell commands
        """
        return RSACommandHandler.SHELL_NAMESPACE

    def get_methods(self):
        """
        Returns the commands provided by this service
        """
        return [
            ("listconfigs", self._list_providers),
            ("lcfgs", self._list_providers),
            ("listproviders", self._list_providers),
            ("listcontainers", self._list_containers),
            ("lcs", self._list_containers),
            ("listexports", self._list_exported_configs),
            ("listimports", self._list_imported_configs),
            ("lexps", self._list_exported_configs),
            ("limps", self._list_imported_configs),
            ("importservice", self._import_edef),
            ("impsvc", self._import_edef),
            ("exportservice", self._export_service),
            ("exportsvc", self._export_service),
            ("unimportservice", self._unimport),
            ("unimpsvc", self._unimport),
            ("unexportservice", self._unexport),
            ("unexpsvc", self._unexport),
            ("showdefaults", self._show_defaults),
            ("setdefaults", self._set_defaults),
            ("showedeffile", self._show_edef),
        ]

    def remote_admin_event(self, event):
        # type: (RemoteServiceAdminEvent) -> None
        """
        Handle a remote service admin event
        """
        if event.get_type() == RemoteServiceAdminEvent.EXPORT_REGISTRATION:
            EDEFWriter().write([event.get_description()], self._edef_filename)

    def _show_defaults(self, io_handler):
        # type: (ShellSession) -> None
        """
        Show default edeffile and default export_config
        """
        io_handler.write_line(
            "Defaults\n\texport_config={0}\n\tEDEF file={1};exists={2}",
            self._export_config,
            self._edef_filename,
            os.path.isfile(self._edef_filename),
        )

    def _set_defaults(self, io_handler, export_config, edef_file=None):
        # type: (ShellSession, str, str) -> None
        """
        Set the export_config and optionally the edef_file default values
        """
        self._export_config = export_config
        if edef_file:
            self._edef_filename = edef_file
        self._show_defaults(io_handler)

    def _show_edef(self, io_handler):
        # type: (ShellSession) -> None
        """
        Show contents of EDEF file
        """
        if not os.path.isfile(self._edef_filename):
            io_handler.write_line(
                "EDEF file '{0}' does not exist!", self._edef_filename
            )
        else:
            with open(self._edef_filename, "r") as f:
                eds = EDEFReader().parse(f.read())

            io_handler.write_line(EDEFWriter().to_string(eds))

    def _list_providers(self, io_handler, provider_id=None):
        # type: (ShellSession, str) -> None
        """
        List export/import providers. If <provider_id> given,
        details on that provider
        """
        with self._bind_lock:
            providers = self._get_dist_providers(provider_id)

        if providers:
            if provider_id:
                provider = providers[0]
                io_handler.write_line(
                    self.PROVIDER_FORMAT.format(
                        provider.get_config_name(),
                        provider.get_supported_configs(),
                        provider.get_supported_intents(),
                    )
                )
            else:
                title = ("Distribution Provider/Config", "Class")
                rows = [
                    (p.get_config_name(), _full_class_name(p))
                    for p in providers
                ]
                io_handler.write_line(self._utils.make_table(title, rows))

    def _list_containers(self, io_handler, container_id=None):
        # type: (ShellSession, str) -> None
        """
        List existing import/export containers.
        If <container_id> given, details on that container
        """
        with self._bind_lock:
            containers = self._get_containers(container_id)
        if containers:
            if container_id:
                container = containers[0]
                connected_id = container.get_connected_id()
                ns = container.get_namespace()
                io_handler.write_line(
                    self.CONTAINER_FORMAT.format(
                        container.get_id(),
                        ns,
                        _full_class_name(container),
                        connected_id,
                        ns,
                        container.get_config_name(),
                    )
                )
            else:
                title = ("Container ID/instance.name", "Class")
                rows = [(c.get_id(), _full_class_name(c)) for c in containers]
                io_handler.write_line(self._utils.make_table(title, rows))

    def _list_imports(self, session, import_regs, endpoint_id=None):
        # type: (ShellSession, List[ImportRegistration], str) -> None
        """
        Lists the imported services
        """
        if endpoint_id:
            matching_eds = [
                x.get_description()
                for x in import_regs
                if x.get_description().get_id() == endpoint_id
            ]
            if matching_eds:
                session.write_line(
                    "Endpoint description for endpoint.id={0}:", endpoint_id
                )
                session.write_line(EDEFWriter().to_string(matching_eds))
        else:
            title = (
                "Endpoint ID",
                "Container ID",
                "Local Service ID",
                "Remote Service ID",
            )
            rows = []
            for import_reg in import_regs:
                ed = import_reg.get_description()
                rows.append(
                    (
                        ed.get_id(),
                        ed.get_container_id()[1],
                        import_reg.get_reference().get_property(SERVICE_ID),
                        ed.get_service_id(),
                    )
                )

            session.write_line(self._utils.make_table(title, rows))

    def _list_exports(self, session, configs, endpoint_id=None):
        # type: (ShellSession, List[ExportRegistration], str) -> None
        """
        Lists the exported services
        """
        if endpoint_id:
            matching_eds = [
                x.get_description()
                for x in configs
                if x.get_description().get_id() == endpoint_id
            ]
            if matching_eds:
                session.write_line(
                    "Endpoint description for endpoint.id={0}:", endpoint_id
                )
                session.write_line(EDEFWriter().to_string(matching_eds))
        else:
            title = ("Endpoint ID", "Container ID", "Service ID")
            rows = []
            for import_reg in configs:
                ed = import_reg.get_description()
                rows.append(
                    (ed.get_id(), ed.get_container_id()[1], ed.get_service_id())
                )

            session.write_line(self._utils.make_table(title, rows))

    def _list_exported_configs(self, io_handler, endpoint_id=None):
        # type: (ShellSession, str) -> None
        # pylint: disable=W0212
        """
        List exported services. If <endpoint_id> given, details on that export
        """
        self._list_exports(
            io_handler, self._rsa._get_export_regs(), endpoint_id
        )

    def _list_imported_configs(self, io_handler, endpoint_id=None):
        # type: (ShellSession, str) -> None
        # pylint: disable=W0212
        """
        List imported endpoints. If <endpoint_id> given, details on that import
        """
        self._list_imports(
            io_handler, self._rsa._get_import_regs(), endpoint_id
        )

    def _unimport(self, io_handler, endpoint_id):
        # type: (ShellSession, str) -> None
        # pylint: disable=W0212
        """
        Un-import endpoint with given endpoint_id (required)
        """
        import_regs = self._rsa._get_import_regs()
        found_reg = None
        for import_reg in import_regs:
            ed = import_reg.get_description()
            if ed and ed.get_id() == endpoint_id:
                found_reg = import_reg
        if not found_reg:
            io_handler.write_line(
                "Cannot find import registration with endpoint.id={0}",
                endpoint_id,
            )
        else:
            # now close it
            found_reg.close()

    def _unexport(self, io_handler, endpoint_id):
        # type: (ShellSession, str) -> None
        """
        Un-export endpoint with given endpoint_id (required)
        """
        # pylint: disable=W0212
        export_regs = self._rsa._get_export_regs()
        found_reg = None
        for export_reg in export_regs:
            ed = export_reg.get_description()
            if ed and ed.get_id() == endpoint_id:
                found_reg = export_reg
        if not found_reg:
            io_handler.write_line(
                "Cannot find export registration with endpoint.id={0}",
                endpoint_id,
            )
        else:
            # now close it
            found_reg.close()

    def _get_edef_fullname(self):
        # type: () -> str
        return self._edef_filename

    def _export_service(
        self, io_handler, service_id, export_config=None, filename=None
    ):
        # type: (ShellSession, str, str, str) -> None
        """
        Export service with given service.id.
        """
        svc_ref = self._context.get_service_reference(
            None, "(service.id={0})".format(service_id)
        )
        if not svc_ref:
            io_handler.write_line(
                "Service with id={0} cannot be found so no service "
                "can be exported",
                service_id,
            )
            return

        if export_config:
            self._export_config = export_config

        if filename:
            self._edef_filename = filename

        # Finally export with required SERVICE_EXPORTED_INTERFACES = '*'
        # and SERVICE_EXPORTED_CONFIGS to self._export_config
        export_regs = self._rsa.export_service(
            svc_ref,
            {
                SERVICE_EXPORTED_INTERFACES: "*",
                SERVICE_EXPORTED_CONFIGS: self._export_config,
            },
        )
        exported_eds = []
        for export_reg in export_regs:
            exp = export_reg.get_exception()
            if exp:
                io_handler.write_line(
                    "\nException exporting service={0}",
                    export_reg.get_reference(),
                )
                print_exception(
                    exp[0], exp[1], exp[2], limit=None, file=io_handler
                )
                io_handler.flush()
            else:
                exported_eds.append(export_reg.get_description())

        # write exported_eds to filename
        EDEFWriter().write(exported_eds, self._edef_filename)

        io_handler.write_line(
            "Service={0} exported by {1} providers. EDEF written to file={2}",
            svc_ref,
            len(exported_eds),
            self._edef_filename,
        )

    def _import_edef(self, io_handler, edef_file=None):
        # type: (ShellSession, str) -> None
        """
        Import endpoint
        """
        if not edef_file:
            edef_file = self._edef_filename

        full_name = self._get_edef_fullname()
        with open(full_name) as f:
            eds = EDEFReader().parse(f.read())
            io_handler.write_line(
                "Imported {0} endpoints from EDEF file={1}", len(eds), full_name
            )

        for ed in eds:
            import_reg = self._rsa.import_service(ed)
            if import_reg:
                exp = import_reg.get_exception()
                ed = import_reg.get_description()
                if exp:
                    io_handler.write_line(
                        "Exception importing endpoint.id={0}", ed.get_id()
                    )
                    print_exception(
                        exp[0], exp[1], exp[2], limit=None, file=io_handler
                    )
                    io_handler.flush()
                else:
                    io_handler.write_line(
                        "Proxy service={0} imported. rsid={1}",
                        import_reg.get_reference(),
                        ed.get_remoteservice_idstr(),
                    )



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

Etcd Discovery Provider

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from threading import Thread, RLock
import logging
import json
import socket
import time

try:
    # pylint: disable=W0611
    from typing import List
except ImportError:
    pass

import etcd

from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Instantiate,
    ValidateComponent,
    Property,
    Invalidate,
)

from pelix.rsa import create_uuid, prop_dot_suffix
from pelix.rsa.endpointdescription import (
    encode_endpoint_props,
    decode_endpoint_props,
    EndpointDescription,
)
from pelix.rsa.providers.discovery import (
    SERVICE_ENDPOINT_ADVERTISER,
    EndpointAdvertiser,
    EndpointEvent,
    EndpointSubscriber,
)

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

ETCD_NAME_PROP = "etcd"
ETCD_HOSTNAME_PROP = "hostname"
ETCD_PORT_PROP = "port"
ETCD_TOPPATH_PROP = "toppath"
ETCD_SESSIONTTL_PROP = "sessionttl"
ETCD_WATCHSTART_WAIT_PROP = "watchstartwait"

# ------------------------------------------------------------------------------


@ComponentFactory("etcd-endpoint-discovery-factory")
@Provides(SERVICE_ENDPOINT_ADVERTISER)
@Property(
    "_hostname",
    prop_dot_suffix(ETCD_NAME_PROP, ETCD_HOSTNAME_PROP),
    "localhost",
)
@Property("_port", prop_dot_suffix(ETCD_NAME_PROP, ETCD_PORT_PROP), 2379)
@Property(
    "_top_path",
    prop_dot_suffix(ETCD_NAME_PROP, ETCD_TOPPATH_PROP),
    "/org.eclipse.ecf.provider.etcd.EtcdDiscoveryContainer",
)
@Property(
    "_session_ttl", prop_dot_suffix(ETCD_NAME_PROP, ETCD_SESSIONTTL_PROP), 30
)
@Property(
    "_watch_start_wait",
    prop_dot_suffix(ETCD_NAME_PROP, ETCD_WATCHSTART_WAIT_PROP),
    5,
)
@Instantiate("etcd-endpoint-discovery")
class EtcdEndpointDiscovery(EndpointAdvertiser, EndpointSubscriber):
    """
    Etcd-based endpoint discovery.  Extends both EndpointAdvertiser
    and EndpointSubscriber so can be called to advertise/unadvertise
    exported endpoints, and will notify SERVICE_ENDPOINT_LISTENERs
    when an endpoint has been discovered via the etcd service.

    Note that this depends upon the python-etcd client library.
    """

    REMOVE_ACTIONS = ["delete", "expire"]
    ADD_ACTIONS = ["set", "create"]

    def __init__(self):
        EndpointAdvertiser.__init__(self)
        EndpointSubscriber.__init__(self)
        self._hostname = self._port = self._top_path = None
        self._sessionid = create_uuid()
        self._session_ttl = self._watch_start_wait = None
        self._client = None  # type: etcd.Client
        self._client_lock = RLock()
        self._top_nodes = (
            self._wait_index
        ) = self._ttl_thread = self._watch_thread = None
        servicename = "osgirsvc_{0}".format(create_uuid())
        hostip = socket.gethostbyname(socket.gethostname())
        self._service_props = {
            "location": "ecfosgisvc://{0}:32565/{1}".format(
                hostip, servicename
            ),
            "priority": 0,
            "weight": 0,
            "servicename": servicename,
            "ttl": 0,
            "servicetype": {
                "services": ["ecfosgirsvc"],
                "scopes": ["default"],
                "protocols": ["default"],
                "namingauth": "iana",
            },
        }

    @ValidateComponent()
    def _validate_component(self):
        # now connect
        self._connect()

    @Invalidate
    def _invalidate(self, _):
        self._disconnect()

    def _encode_description(self, endpoint_description):
        # type: (EndpointDescription) -> dict
        encoded_props = encode_endpoint_props(endpoint_description)
        # get copy of service props
        service_props = self._service_props.copy()
        # set 'properties field'
        service_props["properties"] = [
            {"type": "string", "name": key, "value": encoded_props.get(key)}
            for key in encoded_props
        ]
        return service_props

    def _write_description(self, endpoint_description):
        # type: (EndpointDescription) -> etcd.EtcdResult
        # encode props as string -> string
        service_props = self._encode_description(endpoint_description)
        # dump service_props to json
        props_json = json.dumps(service_props)
        # write to etcd
        with self._client_lock:
            return self._client.write(
                key=self._get_endpoint_path(endpoint_description.get_id()),
                value=props_json,
            )

    # implementation of EndpointAdvertiser service.  These methods
    # are called when (e.g.) RSA asks us to advertise/unadvertise
    # an endpoint_description
    def _advertise(self, endpoint_description):
        # type: (EndpointDescription) -> etcd.EtcdResult
        _logger.debug("advertising ed=%s", endpoint_description)
        return self._write_description(endpoint_description)

    def _update(self, endpoint_description):
        # type: (EndpointDescription) -> etcd.EtcdResult
        _logger.debug("updating ed=%s", endpoint_description)
        return self._write_description(endpoint_description)

    def _unadvertise(self, advertised):
        # type: (List[EndpointDescription]) -> etcd.EtcdResult
        _logger.debug("unadvertising ed=%s", advertised[0])
        # get endpoint id
        endpointid = advertised[0].get_id()
        # write to etcd
        with self._client_lock:
            return self._client.delete(key=self._get_endpoint_path(endpointid))

    def _get_session_path(self):
        # type: () -> str
        return "{0}/{1}".format(self._top_path, self._sessionid)

    def _get_endpoint_path(self, endpointid):
        # type: (str) -> str
        return "{0}/{1}".format(self._get_session_path(), endpointid)

    def _disconnect(self):
        """
        Disconnects the etcd client
        """
        with self._client_lock:
            if self._client:
                session_path = self._get_session_path()
                try:
                    self._client.delete(session_path, True, True)
                except Exception:
                    _logger.exception(
                        "Exception deleting session_path=%s", session_path
                    )
                self._client = None

    def _connect(self):
        """
        Connects to etcd
        """
        with self._client_lock:
            if self._client:
                raise Exception("already connected")
            # create etcd Client instance
            self._client = etcd.Client(host=self._hostname, port=self._port)
            # now make request against basic
            try:
                top_response = self._client.read(self._top_path, recursive=True)
            except etcd.EtcdKeyNotFound:
                # if this happens, attempt to write it
                try:
                    top_response = self._client.write(
                        self._top_path, None, None, True
                    )
                except Exception as e:
                    _logger.exception(
                        "Exception attempting to create top dir=%s",
                        self._top_path,
                    )
                    raise e
            # set top nodes with list comprehension base top_response subtree
            self._top_nodes = [
                x
                for x in list(top_response.get_subtree())
                if x.dir and x.key != self._top_path
            ]
            try:
                session_exists_result = self._client.write(
                    key=self._get_session_path(),
                    value=None,
                    ttl=self._session_ttl,
                    dir=True,
                    prevExist=False,
                )
            except Exception as e:
                _logger.exception(
                    "Exception creating session for client at session_path=%s",
                    self._get_session_path(),
                )
                raise e

            # Note: error disabled as EtcdResult object is too dynamic
            # pylint: disable=E1101
            self._wait_index = session_exists_result.createdIndex + 1
            self._ttl_thread = Thread(target=self._ttl_job, name="Etcd TTL Job")
            self._ttl_thread.daemon = True
            self._watch_thread = Thread(
                target=self._watch_job, name="Etcd Listen Job"
            )
            self._watch_thread.daemon = True
            self._ttl_thread.start()
            self._watch_thread.start()

    def _get_start_wait(self):
        # type: () -> int
        return int(self._session_ttl - (self._session_ttl / 10))

    def _handle_add_dir(self, dir_node):
        sessionid = dir_node.key[len(self._top_path) + 1 :]
        _logger.debug("_handle_add_dir sessionid=%s", sessionid)
        self._handle_add_nodes(
            sessionid,
            [node for node in list(dir_node.children) if not node.dir],
        )

    def _handle_remove_dir(self, sessionid):
        _logger.debug("_handle_remove_dir sessionid=%s", sessionid)
        endpointids = self._get_endpointids_for_sessionid(sessionid)
        self._handle_remove_nodes(endpointids)

    def _handle_add_nodes(self, sessionid, nodes):
        for node in nodes:
            # we only care about properties
            node_val = node.value
            if node_val:
                json_obj = json.loads(node_val)
                if isinstance(json_obj, dict):
                    json_properties = json_obj["properties"]
                    # get the name and value from each entry
                    raw_props = {
                        entry["name"]: entry["value"]
                        for entry in json_properties
                        if entry["type"] == "string"
                    }
                    # decode
                    decoded_props = decode_endpoint_props(raw_props)
                    new_ed = EndpointDescription(properties=decoded_props)
                    old_ed = self._has_discovered_endpoint(new_ed.get_id())
                    if not old_ed:
                        # add discovered endpoint to our internal list
                        self._add_discovered_endpoint(sessionid, new_ed)
                        # dispatch
                        self._fire_endpoint_event(EndpointEvent.ADDED, new_ed)
                    else:
                        # get timestamp and make sure new one is newer (an
                        # update)
                        old_ts = old_ed.get_timestamp()
                        new_ts = new_ed.get_timestamp()
                        if new_ts > old_ts:
                            self._remove_discovered_endpoint(old_ed.get_id())
                            self._add_discovered_endpoint(sessionid, new_ed)
                            self._fire_endpoint_event(
                                EndpointEvent.MODIFIED, new_ed
                            )

    def _handle_remove_nodes(self, endpointids):
        for endpointid in endpointids:
            self._handle_remove_node(endpointid)

    def _handle_remove_node(self, endpointid):
        ed = self._remove_discovered_endpoint(endpointid)
        if ed:
            self._fire_endpoint_event(EndpointEvent.REMOVED, ed)

    def _watch_job(self):
        # sleep for a few seconds to allow endpoint listeners to be
        # asynchronously added before the top nodes are processed
        time.sleep(5)
        # first thing is to process the existing nodes from connect
        if self._top_nodes:
            # guaranteed to be directory
            for dir_node in self._top_nodes:
                self._handle_add_dir(dir_node)
            self._top_nodes = None
        # then loop forever
        while True:
            with self._client_lock:
                client = self._client
            if not client:
                return
            try:
                result = client.read(
                    key=self._top_path,
                    recursive=True,
                    wait=True,
                    waitIndex=self._wait_index,
                )

                # reset wait_index
                # Note: error disabled as EtcdResult object is too dynamic
                # pylint: disable=E1101
                self._wait_index = result.modifiedIndex + 1
                key = result.key
                action = result.action
                if key.endswith(self._sessionid):
                    if action == "delete":
                        _logger.debug(
                            "watch_job: session dir deleted...exiting"
                        )
                        # we are done
                        return
                else:
                    # split id into [sessionid] or [sessionid,endpointid]
                    splitid = key[len(self._top_path) + 1 :].split("/")
                    sessionid = splitid[0]
                    # only process sessionids that are not ours
                    if self._sessionid != sessionid:
                        # if length of splitid list is > 1 then it's a leaf
                        # node
                        if len(splitid) > 1:
                            endpointid = splitid[len(splitid) - 1]
                            if action in self.REMOVE_ACTIONS:
                                self._handle_remove_node(endpointid)
                            elif action in self.ADD_ACTIONS:
                                self._handle_add_nodes(sessionid, [result])
                        # otherwise it's a branch/dir node
                        else:
                            if action in self.REMOVE_ACTIONS:
                                self._handle_remove_dir(sessionid)
                            elif action in self.ADD_ACTIONS:
                                self._handle_add_dir(result)

            except Exception:
                _logger.exception("watch_job:Exception in watch loop")

    def _ttl_job(self):
        waittime = self._get_start_wait()
        while True:
            _logger.debug("ttl_job: starting sleep with waittime=%s", waittime)
            time.sleep(1)
            with self._client_lock:
                client = self._client
            if not client:
                _logger.debug("ttl_job: exiting")
                return
            waittime -= 1
            _logger.debug(
                "ttl_job: testing waittime <= 0 with waittime=%s", waittime
            )
            if waittime <= 0:
                try:
                    session_ttl = self._session_ttl
                    _logger.debug(
                        "ttl_job: updating with session_ttl=%s", session_ttl
                    )
                    with self._client_lock:
                        if self._client:
                            self._client.write(
                                key=self._get_session_path(),
                                value=None,
                                ttl=session_ttl,
                                dir=True,
                                prevExist=True,
                            )
                    _logger.debug(
                        "ttl_job: updated with session_ttl=%s", session_ttl
                    )
                except Exception:
                    _logger.exception("Exception updating in ttl job")
                waittime = self._get_start_wait()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
@RequiresVarFilter Dependency handler

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import string

# Pelix
import pelix.ldapfilter as ldapfilter
from pelix.constants import BundleActivator

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants
import pelix.ipopo.handlers.requires as requires

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(requires._HandlerFactory):
    # pylint: disable=R0903, W0212
    """
    Factory service for service registration handlers
    """

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Extract information from the context
        requirements = component_context.get_handler(
            ipopo_constants.HANDLER_REQUIRES_VARIABLE_FILTER
        )
        requires_filters = component_context.properties.get(
            ipopo_constants.IPOPO_REQUIRES_FILTERS, None
        )

        # Prepare requirements
        requirements = self._prepare_requirements(
            requirements, requires_filters
        )

        # Set up the runtime dependency handlers
        handlers = []
        for field, requirement in requirements.items():
            # Construct the handler
            if requirement.aggregate:
                handlers.append(
                    AggregateDependency(component_context, field, requirement)
                )
            else:
                handlers.append(
                    SimpleDependency(component_context, field, requirement)
                )

        return handlers


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_REQUIRES_VARIABLE_FILTER
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class _VariableFilterMixIn:
    """
    Dependency handler MixIn to support variable filters
    """

    def __init__(self, component_context, requirement):
        """
        Set up the MixIn
        """
        # Component context
        self._component_context = component_context

        # Copy the current filter as a string
        self._original_filter = str(requirement.filter)
        self.valid_filter = False

        # List the properties found in the filter
        self._keys = self._find_keys()

        try:
            # Set the initial value of the filter
            self.update_filter()
        except ValueError:
            # The filter couldn't be initialized (reason already logged)
            self.valid_filter = False

    def _find_keys(self):
        """
        Looks for the property keys in the filter string

        :return: A list of property keys
        """
        formatter = string.Formatter()
        return [
            val[1] for val in formatter.parse(self._original_filter) if val[1]
        ]

    def update_filter(self):
        """
        Update the filter according to the new properties

        :return: True if the filter changed, else False
        :raise ValueError: The filter is invalid
        """
        # Consider the filter invalid
        self.valid_filter = False

        try:
            # Format the new filter
            filter_str = self._original_filter.format(
                **self._component_context.properties
            )
        except KeyError as ex:
            # An entry is missing: abandon
            logging.warning("Missing filter value: %s", ex)
            raise ValueError("Missing filter value")

        try:
            # Parse the new LDAP filter
            new_filter = ldapfilter.get_ldap_filter(filter_str)
        except (TypeError, ValueError) as ex:
            logging.warning("Error parsing filter: %s", ex)
            raise ValueError("Error parsing filter")

        # The filter is valid
        self.valid_filter = True

        # Compare to the "old" one
        if new_filter != self.requirement.filter:
            # Replace the requirement filter
            self.requirement.filter = new_filter
            return True

        # Same filter
        return False

    def on_property_change(self, name, old_value, new_value):
        # pylint: disable=W0613
        """
        A component property has been updated

        :param name: Name of the property
        :param old_value: Previous value of the property
        :param new_value: New value of the property
        """
        if name in self._keys:
            try:
                if self.update_filter():
                    # This is a key for the filter and the filter has changed
                    # => Force the handler to update its dependency
                    self._reset()
            except ValueError:
                # Invalid filter: clear all references, this will invalidate
                # the component
                for svc_ref in self.get_bindings():
                    self.on_service_departure(svc_ref)

    def _reset(self):
        """
        Called when the filter has been changed
        """
        with self._lock:
            # Start listening to services with the new filter
            self.stop()
            self.start()

            # Force bindings update
            self._ipopo_instance.update_bindings()

            for svc_ref in self.get_bindings():
                # Check if the current reference matches the filter
                if not self.requirement.filter.matches(
                    svc_ref.get_properties()
                ):
                    # Not the case: emulate a service departure
                    # The instance life cycle will be updated as well
                    self.on_service_departure(svc_ref)


class SimpleDependency(_VariableFilterMixIn, requires.SimpleDependency):
    """
    Manages a single dependency field
    """

    def __init__(self, component_context, field, requirement):
        """
        Sets up members
        """
        requires.SimpleDependency.__init__(self, field, requirement)
        _VariableFilterMixIn.__init__(self, component_context, requirement)

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return self.valid_filter and requires.SimpleDependency.is_valid(self)

    def try_binding(self):
        """
        Searches for the required service if needed

        :raise BundleException: Invalid ServiceReference found
        """
        if self.valid_filter:
            # Look for a service only if the filter is valid
            requires.SimpleDependency.try_binding(self)


class AggregateDependency(_VariableFilterMixIn, requires.AggregateDependency):
    """
    Manages a single dependency field
    """

    def __init__(self, component_context, field, requirement):
        """
        Sets up members
        """
        requires.AggregateDependency.__init__(self, field, requirement)
        _VariableFilterMixIn.__init__(self, component_context, requirement)

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return self.valid_filter and requires.AggregateDependency.is_valid(self)

    def try_binding(self):
        """
        Searches for the required service if needed

        :raise BundleException: Invalid ServiceReference found
        """
        if self.valid_filter:
            # Look for a service only if the filter is valid
            requires.AggregateDependency.try_binding(self)
